#!/usr/bin/env python
""" PIPE STEP PARENT - Version 1.0.0

    This module defines the pipeline step parent object. Pipe steps are
    the modules responsible for all data reduction. They are called by
    the pipeline and work with pipedata objects. All pipe step objects are
    descendants from stepparent. Pipe steps are callable objects that return
    the reduced data product (as pipedata object).
    
    @author: berthoud
"""

import logging # logging object library
from darepype.drp.stepparent import StepParent

class blah(StepParent):
    """ HAWC Pipeline Step Parent Object
        The object is callable. It requires a valid configuration input
        (file or object) when it runs.
    """
    stepver = '0.1' # pipe step version
    
    def setup(self):
        """ ### Names and Parameters need to be Set Here ###
            Sets the internal names for the function and for saved files.
            Defines the input parameters for the current pipe step.
            Setup() is called at the end of __init__
            The parameters are stored in a list containing the following
            information:
            - name: The name for the parameter. This name is used when
                    calling the pipe step from command line or python shell.
                    It is also used to identify the parameter in the pipeline
                    configuration file.
            - default: A default value for the parameter. If nothing, set
                       '' for strings, 0 for integers and 0.0 for floats
            - help: A short description of the parameter.
        """
        ### Set Names
        # Name of the pipeline reduction step
        self.name='parent'
        # Shortcut for pipeline reduction step and identifier for
        # saved file names.
        self.procname = 'unk'
        # Set Logger for this pipe step
        self.log = logging.getLogger('pipe.step.%s' % self.name)
        ### Set Parameter list
        # Clear Parameter list
        self.paramlist = []
        # Append parameters
        self.paramlist.append(['sampar', 1.0,
            'Sample Parameter - parent only - no practical use'])

    def run(self):
        """ Runs the data reduction algorithm. The self.datain is run
            through the code, the result is in self.dataout.
        """
        self.log.debug('Running step %s' % self.name)
    
    def reset(self):
        """ Resets the step to the same condition as it was when it was
            created. Internal variables are reset, any stored data is
            erased.
        """
        self.log.debug('Reset: done')
        
    def test(self):
        """ Test Pipe Step Parent Object:
            Runs a set of basic tests on the object
        """
        # log message
        self.log.info('Testing pipe step %s' %self.name)

        # log message
        self.log.info('Testing pipe step %s - Done' %self.name)
    
if __name__ == '__main__':
    """ Main function to run the pipe step from command line on a file.
        Command:
          python stepparent.py input.fits -arg1 -arg2 . . .
        Standard arguments:
          --config=ConfigFilePathName.txt : name of the configuration file
          -t, --test : runs the functionality test i.e. pipestep.test()
          --loglevel=LEVEL : configures the logging output for a particular level
          -h, --help : Returns a list of 
    """
    blah().execute()

""" === History ===
"""



''' PIPE STEP Multi-Output PARENT

    This module defines a pipeline step with multiple input and output data
    objects. It is a child object of StepMIParent. All such pipe steps are
    descendants of this one.

    @author: berthoud
'''

import logging # logging object library
from .dataparent import DataParent # pipeline data object
from .stepmiparent import StepMIParent # pipe step parent object

class StepMOParent(StepMIParent):
    """ Pipeline multiple data input and output parent object
    """
    
    stepver = '1.0' # pipe step version
    
    def __init__(self):
        """ Constructor: Initialize data objects and variables
            calls the setup function.
        """
        # call superclass constructor (calls setup)
        super(StepMOParent,self).__init__()
        # Change dataout
        self.dataout = [DataParent()]
        # set iomode
        self.iomode = 'MIMO'

    def setup(self):
        """ ### Names and Prameters need to be Set Here ###
            Sets the internal names for the function and for saved files.
            Defines the input parameters for the current pipe step.
            Setup() is called at the end of __init__
            The parameters are stored in a list containing the following
            information:
            - name: The name for the parameter. This name is used when
                    calling the pipe step from command line or python shell.
                    It is also used to identify the parameter in the pipeline
                    configuration file.
            - default: A default value for the parameter. If nothing, set
                       '' for strings, 0 for integers and 0.0 for floats
            - help: A short description of the parameter.
        """
        ### Set Names
        # Name of the pipeline reduction step
        self.name='parentmo'
        # Shortcut for pipeline reduction step and identifier for
        # saved file names.
        self.procname = 'unk'
        # Set Logger for this pipe step
        self.log = logging.getLogger('pipe.step.%s' % self.name)
        ### Set Parameter list
        # Clear Parameter list
        self.paramlist = []
        # Append parameters
        self.paramlist.append(['sampar', 1.0,
            'Sample Parameter - parent only - no practical use'])

    def run(self):
        """ Runs the data reduction algorithm. The self.datain is run
            through the code, the result is in self.dataout.
        """
        # Log the value of sample parameter
        self.log.debug("Sample Parameter = %.2f" % self.getarg('sampar'))
        # Return the first datain element
        self.dataout = self.datain

    def runend(self,data):
        """ Method to call at the end of pipe the pipe step call
           - Sends final log messages
        """
        # update header (status and history)
        for d in data:
            self.updateheader(d)
        # clear input arguments
        self.arglist = {}
        self.log.info('Finished Reduction: Pipe Step %s' % self.name)

    def execfiles(self, inputfiles):
        """ Runs several files from execute simultaneously.
        """
        if len(inputfiles) > 0:
            # Read input files to datain
            self.datain = []
            for filename in inputfiles:
                # Read input file
                data = DataParent(config = self.config)
                self.datain.append(data.load(filename))
            # Call start - run and call end
            self.runstart(self.datain,self.arglist)
            self.run()
            self.runend(self.dataout)
            # Write output files
            for data in self.dataout:
                data.save()
                self.log.info('Execute: Saved result %s' % data.filename)
        else:
            # Warning - no input file(s)
            self.log.warn('Execute: Missing input File(s)')
    
    def test(self):
        """ Test Pipe Step Parent Object:
            Runs a set of basic tests on the object
        """
        # log message
        self.log.info('Testing pipe step %s' %self.name)
        # read configuration
        if self.config != None:
            datain = DataParent(config=self.config)
        else:
            datain = DataParent(config=self.testconf)
        # generate 2 files
        datain.filename = 'this.file.type.fts'
        datain = [datain,datain]
        # run function call
        dataout = self(datain)      
        # test output
        print(self.config[self.name])
        print(type(dataout))
        print(dataout[0].header)
        # log message
        self.log.info('Testing pipe step %s - Done' %self.name)


if __name__ == '__main__':
    """ Main function to run the pipe step from command line on a file.
        Command:
          python stepparent.py input.fits -arg1 -arg2 . . .
        Standard arguments:
          --config=ConfigFilePathName.txt : name of the configuration file
          -t, --test : runs the functionality test i.e. pipestep.test()
          --loglevel=LEVEL : configures the logging output for a particular level
          -h, --help : Returns a list of 
    """
    StepMOParent().execute()

""" === History ===
    2014-3-19 Marc Berthoud: Written and tested version 0.1
"""



#! /usr/bin/env python

""" PIPE LINE - Version 1.0.0
    This module defines the HAWC data reduction pipeline object. The
    object identifies, loads, reduces and stores raw HAWC data.
"""

import os # os library
import logging # logging object library
import traceback # error traceback library
import argparse # Argument parsing library
import configobj # Config Object library
from .dataparent import DataParent # Pipeline Data object
from .stepparent import StepParent # Step Parent
from .stepmiparent import StepMIParent # Step Multiple Input Parent
from .stepniparent import StepNIParent # Step No Input Parent

class PipeLine(object):
    """ HAWC Pipeline Object
        The object receives filesnames or loaded pipedata objects,
        reduces them and adds them to the current data. The pipeline object
        identifies the instrument mode and loads the correct configuration
        file. The configuration file also contains information which pipe
        steps to use.
    """

    def __init__(self, filenames=None, config=None, pipemode=None,
                 runmode=None, savemem=False):
        """ Constructor: Initialize the pipeline
        """
        # Admin Variables
        self.pipemode = '' # pipe mode: determines which mode the pipeline
                            # is running, i.e. which steps are run
        self.runmode = 'auto' # pipe running modes: AUTOmatic or INTERactive
        self.errstop = 0 # flag for pipe to stop if pipe step returns error
        self.config = None # configuration object
        self.memFlag = savemem # set to True to conserve memory.  This will
                                # not save outputs of each step, and alter
                                # the behaviour of self.getresults()
        self.log = logging.getLogger('pipe.line')
        # Pipesteps Variables
        self.stepnames = [] # names of the pipeline steps (from step.name)
        self.steps = [] # list with the pipeline step objects
                        # ( steps[i] is of type stepname[i] )
                        # ( can't use dict b/c dicts have no order )
                        ### len(self.steps) > 0 indicates pipe is set up
        # Data Variables
        self.results = [] # list with results from each file for the most recent
                            #     step (Can be pipedata objects or lists thereof)
        self.outputs = [] # list of most recent outputs for each step
        self.inputs = [] # list of input data, one element for each step
                        # these are only used for Multiple Input steps
        self.finals = [] # list of the results from last few final steps
        # Files Variables
        self.openfiles = [] # file name list of non-reduced files (fifo)
        self.reducedfiles = [] # file name list of reduced files
        self.outfiles = [] # file name list of files saved to disk
        # set up the pipeline
        self.setup(filenames=filenames, config=config, pipemode=pipemode,
                   runmode=runmode)
        # message
        self.log.debug('Init - Done')

    def reset(self):
        """ Resets the pipeline and prepares it to reduce a new set of
            images. Any reduced results will be deleted (make sure you save
            before you clear). All pipestep objects will be destroyed.
        """
        # Reset settings
        self.pipemode = ''
        # Reset pipeline
        for step in self.steps:
            del step
        self.stepnames=[]
        self.steps=[]
        self.openfiles=[]
        self.reducedfiles=[]
        self.outfiles=[]
        # Reset Data
        self.results=[]
        self.finals=[]
        # message
        self.log.debug('Cleared - Done')

    def setup(self, filenames=None, config=None, pipemode=None,
              runmode=None, errstop=None):
        """ Using the information in the (first) file in 'filenames', this
            function sets up the pipeline. If pipemode is None, the header
            keywords in the file are used to determine pipemode.
            Also sets configuration for the pipeline: if a filename is
            specified, the configuration file is read. The configuration
            object is returned.
            The configuration can always be changed, but once the pipemode and
            the pipe steps have been set they can only be changed by using the
            reset() command, followed by a new setup().
        """
        ### set runmode
        if runmode ==  'auto' or runmode == 'inter':
            self.runmode=runmode
        elif runmode != None:
            self.log.warn('Setup: Invalid Runmode (%s)' % runmode)
        ### set configuration
        if config != None:
            self.config = DataParent().setconfig(config)
        ### set errstop from input or from config
        if errstop != None:
            self.errstop = errstop
        else:
            try: # necessary because various errors are possible
                self.errstop = int(self.config['general']['errorstop'] )
            except:
                pass
        ### get file header
        if filenames == None or len(filenames) == 0:
            # no file -> say it
            headdata = None
            retmsg='no FITS files given'
        else:
            # try to get the length of filenames[0]
            try:
                namelen=len(filenames[0])
            except TypeError as error:
                self.log.error('Setup: invalid filenames type')
                raise error
            except IndexError as error:
                self.log.error('Setup: no input files')
                raise error
            # if len(filenames[0]) > 1 it should be a filename (from a list)
            if namelen > 1:
                firstfile = filenames[0]
            else:
                firstfile = filenames
            # load header
            headdata = DataParent(config=self.config).loadhead(firstfile)
            retmsg='Loaded file header'
        ### get pipeline mode
        # If available: from input parameter
        if pipemode != None:
            if len(self.pipemode):
                if self.pipemode != pipemode:
                    self.log.warn('Attempt to overwrite current PIPEMODE ('
                                  +self.pipemode+') with '+pipemode
                                  +' - ignoring')
            else:
                self.pipemode = pipemode
        # else: get pipeline mode from file header
        elif headdata != None and len(self.pipemode) == 0 :
            self.pipemode = self.getpipemode(headdata)
            if len(self.pipemode) == 0:
                msg = "Setup: No Pipe Mode found matching initial data"
                self.log.error(msg)
                raise KeyError(msg)
        retmsg=retmsg+' / PipeMode=%s' % self.pipemode
        ### read steps (if config and pipemode available and no steps loaded)
        if ( len(self.pipemode) > 0 and self.config != None and 
             len(self.steps) == 0 ):
            self.log.info('Pipe Mode = %s' %self.pipemode)
            conf_section = 'mode_' + self.pipemode
            # Update configuration based on mode preferences
            confname = self.config.filename
            self.config = configobj.ConfigObj(self.config) # make copy to preserve orig
            self.config.filename = confname
            modeconf = self.config[conf_section]
            for item in modeconf.sections:
                for subitem in modeconf[item]:
                    self.config[item][subitem]=modeconf[item][subitem]
                    self.log.debug('Setup: change conf[%s][%s] to %s' %
                                   (item,subitem,modeconf[item][subitem]))
            # get list of steps - copy to steplist
            try:
                steplist = list(self.config[conf_section]['stepslist'])
            except KeyError as error:
                self.log.error('Setup: Missing steplist item in configuration '
                               +' for PipeMode = ' + self.pipemode )
                raise error
            self.stepnames = steplist
            # initialize inputs and outputs
            self.inputs = []
            self.outputs = []
            # get the steps - loop through stepnames
            dp = DataParent(config = self.config)
            for stepi in range(len(self.stepnames)):
                stepname = self.stepnames[stepi]
                # Add step to steps
                if stepname in 'load save' or stepname[:5] == 'load_':
                    self.steps.append(stepname)
                else:
                    # Use PipeData.getobject to get step object
                    stepobj = dp.getobject(stepname)
                    self.steps.append(stepobj)
                    # add real stepnames from step.name, fill self.inputs
                    self.stepnames[stepi]=self.steps[stepi].name
                self.inputs.append([])
                self.outputs.append(None)
            retmsg=retmsg+' / Set up %s pipe steps' % steplist
        self.log.debug('Setup: done - %s' % retmsg)

    def getpipemode(self,data):
        """ Searches for an appropriate pipe mode in the config file, given
            the header values in the passed data. Tries to mach all key=value
            pairs in the datakeys value of the mode_ entries in the config
            file.  Returns name of first drp mode that matches
            the data.  Returns ' ' if no matching drp mode found.  
        """
        for section in self.config.sections:
            if section.startswith('mode_'):
                # Get the datakeys and make list of lists with
                # format [ [key, val], [key, val], [key,val] ]
                try:
                    datakeys = self.config[section]['datakeys'].split('|')
                except KeyError:
                    self.log.warning("In configuration, missing datakeys for mode=%s"
                                     % section)
                    continue
                datakeys = [dk.split('=') for dk in datakeys]
                # Check all keywords in the file
                check = True
                for dk in datakeys:
                    try:
                        value = str(data.getheadval(dk[0].strip(), errmsg=False))
                        if value.upper().strip() != dk[1].upper().strip():
                            check = False
                    except KeyError:
                        self.log.debug('GetPipeMode: Data is missing key=%s', dk[0])
                        check = False
                if check:
                    self.log.debug('GetPipeMode: Found mode=%s' % section[5:])
                    return section[5:] # return mode name w/o 'mode_'
        # should only get here in no matching drp mode found
        msg = "GetPipeMode: No DRP mode found matching data = %s" % data.filename
        self.log.warning(msg)
        return ''

    def getsteps(self):
        """ Returns the list of the pipeline steps in this pipeline.
            If the list of steps is not yet known, an empty list is
            returned.
        """
        self.log.debug('GetSteps: done')
        return self.steps

    def getresult(self, stepname='final'):
        """ Returns the last result from the 'stepname' pipeline reduction
            step. Returns null if stepname is invalid or if that reduction
            step has not yet been run. If stepname is 'final' the result from
            the last step is returned. If stepname is 'finals' the list of
            all the final results is returned.
        """
        # get index
        if stepname.lower() == 'final':
            resind = len(self.steps)-1
        else:
            try:
                resind = self.stepnames.index(stepname.lower())
            except ValueError as error:
                self.log.error("GetResult: invalid step <%s> requested"
                               % stepname)
                raise error
        # check if data is available
        if len(self.reducedfiles) < 1:
            self.log.warning('GetResult: no results yet')
            return 0
        # get result
        if self.memFlag: # memory conserved, so returns self.results
            self.log.warning('GetResult: Memory conserved, so result of last step returned')
            ret = self.results[-1]
        else:
            ret = self.outputs[resind]
        if isinstance(ret, (list,tuple)):
            ret = ret[0]
        # return result
        self.log.debug('GetResult: done')
        return ret

    def addfiles(self, filenames):
        """ Prepares the listed file(s) for reduction. If no files
            have been loaded, the instrument mode is determined from the
            first file and the corresponding configuration file is loaded.
            The files have to be of the correct mode.
        """
        # try to get the length of filenames[0]
        try:
            namelen=len(filenames[0])
        except TypeError as error:
            self.log.error('Add: invalid filenames type')
            raise error
        except IndexError as error:
            # If filenames is empty -> just move along
            namelen = 2
        # make sure filenames is a valid list
        if namelen < 2: filenames=[filenames]
        # add the files to the list (check if each file exists)
        for filename in filenames:
            if os.path.isfile(filename):
                self.openfiles.append(filename)
            else:
                self.log.warn('Add: could not find file (%s) - ignored'
                              % filename)
        self.log.debug('Add - Done')

    def save(self, filename=None, stepname='final'):
        """ Saves the result from the specified pipeline step under filename.
            If an error is raised if the step 'stepname' has not yet been
            run or if stepname is invalid. If stepname is 'final', the result
            from the last step is saved.
        """
        result = self.getresult(stepname=stepname)
        if filename is None:
            filename = result.filename
        result.save(filename)
        self.outfiles.append(filename)
        self.log.debug('Save - Done')

    def __call__(self, filenames = None, config = None, pipemode = None,
                 runmode = None, force=False):
        """ Runs the pipeline with the current files.

            Files with instmode/instcfg that do not match the pipeline's
            instmode/instcfg, are skipped.  This behavior can be overwritten
            by setting force = True (i.e. pipeline will attempt non-matching
            files anway).
        """
        ### Setting up
        # Add files to file list
        if filenames != None:
            self.addfiles(filenames)
        # Configure
        self.setup(filenames=self.openfiles, config=config, pipemode=pipemode,
                   runmode=runmode)
        # check if steps are available
        if len(self.steps) == 0 :
            self.log.error('Call: pipeline is not set up - aborting')
            raise RuntimeError('Call: pipeline is not set up - aborting')
        ### Initialize results:
        #   - make a pipedata object for each file with filename and header
        #   - check if it's a valid file (instmode)
        # Reset resuls
        self.results = []
        # Loop through files in openfiles
        for filename in self.openfiles:
            # Load file header in new data object
            data = DataParent(config=self.config).loadhead(filename)
            fileshort = os.path.split(filename)[-1]
            self.log.info('Preparing file %s' % fileshort)
            # Check if instmode and instcfg match.  If not, skip file
            filepipemode = self.getpipemode(data)
            if self.pipemode != filepipemode:
                if force is True:
                    self.log.warning('Call: File %s does not fit PipeMode datakeys  - Will attempt to process' %filename)
                else:
                    self.log.warning('Call: File %s does not fit PipeMode datakeys - skipping' %filename)
                    self.openfiles.remove(filename)
                    continue
            # Add file to results
            self.results.append(data)
            # Add to reducedfiles
            self.reducedfiles.append(filename)
        # Remove the files from openfiles
        self.openfiles = []
        ### Loop over steps:
        if len(self.results) == 0 and not isinstance(self.steps[0], StepNIParent):
            self.log.error('Call: Zero files to process - quitting')
            stepi = len(self.steps)
        else:
            stepi = 0
        while stepi < len(self.steps):
            # Get stepstart and stepend i.e. range of next SISO steps
            # Point stepend to first MI step after stepstart or last step
            stepstart = stepi
            nextMI = False # flag indicating next step is MI
            stepend = stepi
            while not nextMI and stepend < len(self.steps) :
                if isinstance(self.steps[stepend], StepMIParent):
                    nextMI = True
                else:
                    stepend += 1
            ### run all SISO steps for each file in results
            #   files with errors are removed from results
            filei = 0
            self.log.debug('Call: Running steps %d to %d for all %d files'
                           % (stepstart, stepend-1, len(self.results)))
            while filei < len(self.results):
                errorflag = False
                stepi = stepstart
                while stepi < stepend:
                    data = self.results[filei]
                    # catch special steps (load/save)
                    if self.stepnames[stepi] == 'load':
                        # Load File
                        data.load()
                        if not self.memFlag:
                            self.outputs[stepi] = data
                    elif self.stepnames[stepi] == 'save':
                        # Save File
                        # Check to make sure pipemode is in history
                        msg = "PipeMode = " + self.pipemode
                        found = False
                        if 'HISTORY' in data.header:
                            for h in data.header['HISTORY']:
                                if msg in h: found = True
                        if not found: data.header['HISTORY'] = msg
                        # Save the file
                        data.save()
                        self.outfiles.append(data.filename)
                        if not self.memFlag:
                            self.outputs[stepi] = data
                    elif self.stepnames[stepi][:5] == 'load_':
                        # Load particular process result (has to be saved first)
                        procname = self.stepnames[stepi][5:].upper()
                        self.log.debug('Loading %s process result' % procname)
                        newfname = data.filenamebegin + procname + data.filenameend
                        data.load(newfname)
                        if not self.memFlag:
                            self.outputs[stepi] = data
                    # run step[stepi]
                    else:
                        self.log.debug('Call: Starting step %s on file %s' %
                                       ( self.stepnames[stepi], data.filename) )
                        if self.errstop > 0:
                            # if errors are not caught -> run the step
                            data = self.steps[stepi](data)
                        else:
                            # try the step catch and handle error
                            try:
                                data = self.steps[stepi](data)
                            except (ValueError, RuntimeError, TypeError,
                                    KeyError, IndexError, IOError) as error:
                                # list warning
                                self.log.warn('Call: Step %s'  % self.stepnames[stepi]
                                    + ' for file %s' % data.filename + ' returned an error,'
                                    + ' message=%s' % str(error) + ' - skipping file')
                                # print traceback
                                traceback.print_exc()
                                # abort this file
                                errorflag = True
                                break # break out of loop over steps
                        self.steps[stepi].datain = None
                        self.steps[stepi].dataout = None
                    # store step output in outputs[stepi]
                    if not self.memFlag:
                        self.outputs[stepi] = data
                    self.results[filei] = data
                    stepi += 1
                # -- end of loop over steps between stepstart and stepend
                # If there's an error in file:
                if errorflag:
                    # Remove file from results forward stepi
                    self.results.remove(self.results[filei])
                    stepi = stepend
                else:
                    # set index for next file
                    filei += 1
            # -- end of loop over files
            ### Here would be a good moment to stop reduction and inquire
            #   If more files arrived to reduce (in that case restart the process)
            ### run an MI step if there is one
            if stepi < len(self.steps) :
                self.log.debug('Call: Running multi-input step %d (%s)' %
                               ( stepi, self.stepnames[stepi]) )
                # add current results to self.inputs[stepi]
                #     overwrite existing entries with identical filenames
                inpfilenames = [inp.filename for inp in self.inputs[stepi]]
                for res in self.results:
                    if res.filename in inpfilenames:
                        n = inpfilenames.index(res.filename)
                        self.inputs[stepi][n] = res
                    else:
                        self.inputs[stepi].append(res)
                # If the step is an NI, set the correct config
                if isinstance(self.steps[stepi], StepNIParent):
                    self.steps[stepi].config = self.config
                # Run the next MI step with the inputs
                try:
                    data = self.steps[stepi](self.inputs[stepi])
                # if there's an error. stop and return an error (no matter what)
                #     make sure results = []
                except (ValueError, RuntimeError, TypeError,
                        KeyError, IndexError, IOError) as error:
                    # list error
                    self.log.error('Call: Step %s'  % self.stepnames[stepi]
                                   + ' for file %s' % self.inputs[stepi][0].filename + ' returned an error,'
                                   + ' message=%s' %str(error) + ' - skipping file')
                    # print traceback
                    traceback.print_exc()
                    # raise error
                    raise(error)
                #self.steps[stepi].reset()
                self.steps[stepi].datain = None
                self.steps[stepi].dataout = None
                # store results in outputs[stepi] and results
                if not self.memFlag:
                    self.outputs[stepi] = data
                if isinstance(data, (list,tuple)):
                    self.results = data
                else:
                    self.results = [data]
                # increase step index
                stepi += 1
        # -- end loop over all steps
        # add output from last step to self.finals
        if len(self.results) > 0:
            self.finals.append(self.results[-1])
        ### Return Result
        self.log.debug('Call: Finished Reducing Queued Data')
        return self.getresult()

    def test(self):
        """ Test PipeLine Object
        """
        # log message
        self.log.info('Testing Pipe Line')
        # set setup
        config = StepParent().testconf
        self.setup(pipemode='chop',config=config,
                   runmode='auto', errstop = 1)
        #self.setup(filenames='080211_000_00Ha010.dog.raw.dogscan.fits',
        #self.setup(filenames='short.fft.raw.fits',
        #           config='pipe_test_conf.txt')
        # get results and test
        #print self.getsteps()
        #print self.getresult('demod')
        # add the files
        print('Reduction Steps = %s' % (self.stepnames,))
        #infiles=['mode_fft/short.fft.raw.fits']
        infiles = ['120306_000_00HA005.chop.raw.fits',
                   '120306_000_00HA006.chop.raw.fits']
        infiles = [os.path.join(self.config['testing']['testpath'],
                                'hawc1G/mode_chop', ifile) for ifile in infiles]
        self.addfiles(infiles)
        # reduce the files
        result = self()
        print('Shape of final Image = %s' % (result.image.shape,))
        # log message
        self.log.info('Testing Pipe Line - Done')


    def execute(self):
        """ Runs the pipeline as called from the command line:
            The first argument is used as configuration file name.
            Following arguments are used as input files. Other
            special arguments are:
            - test: runs the test function
            - loglevel: name of logging level (INFO is default)
            - logfile: log file name
        """
        ### Read Arguments
        # Set up argument parser - Generic parameters
        parser = argparse.ArgumentParser(description="Pipe Line",
            epilog = 'Delta config files specified by --config are merged with baseconfig.')
        parser.add_argument('baseconfig', default = 'pipeconf.txt', type=str, nargs='?',
                            help='pipeline configuration file (required, default = pipeconf.txt)')
        parser.add_argument('inputfiles', type=str, default='', nargs='*',
                            help='input files pathname',)
        parser.add_argument('-t','--test', action='store_true',
                            help='runs the selftest of the pipeline')
        parser.add_argument('--loglevel', default='INFO', type=str,
                            choices=['DEBUG','INFO','WARN',
                                     'ERROR','CRITICAL'],
                            help='log level (default = INFO)')
        parser.add_argument('--logfile', default='', type=str,
                            help='logging file (default = none)')
        parser.add_argument('--pipemode', default='', type=str,
                            help='pipeline mode (default = none)')
        parser.add_argument('-c', '--config', default=[], type=str,
                            action='append', help='Delta configuration file(s) (optional)')
        # Get arguments - store dict in arglist
        args = parser.parse_args()
        self.arglist=vars(args)
        ### Process generic arguments
        # Set logging (add file handler if logfile != '')
        level = getattr(logging, args.loglevel.upper(), None)
        logging.basicConfig(level=level)
        if len(args.logfile) > 0:
            fhand = logging.FileHandler(args.logfile)
            fmt = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
            fhand.setFormatter(logging.Formatter(fmt))
            logging.getLogger().addHandler(fhand)
        # Check for test
        if args.test:
            self.test()
            return
        # Get pipemode
        pipemode = None
        force = False
        if len(args.pipemode) > 0:
            pipemode = args.pipemode
            force = True
        # Set configuration
        config = [args.baseconfig] + args.config
        self.config = DataParent(config = config).config
        ### Reduce data
        if len(args.inputfiles) > -1:
            self(args.inputfiles,pipemode=pipemode,force=force)
            self.save()
        else:
            # Warning - no input file
            self.log.warn('Execute: Missing input File')
        self.log.info('Execute: done')


if __name__ == '__main__':
    """ Main function to run the pipe line from command line on a file.
        Command:
          python stepparent.py pipeconf.txt input1.fits input2.fits -arg1 -arg2 . . .
        where pipeconf.txt stands for the (path)name of the pipeline
        configuration file and inputN.fits for the input file(s).

        Standard arguments:
          -t, --test : runs the functionality test i.e. pipestep.test()
          --loglevel=LEVEL : configures the logging output for a particular level
          -h, --help : Returns a list of commands.
    """
    PipeLine().execute()

""" === History ===
    2016-4-14 Marc Berthoud: Added ability to specify configuration changes in
               mode prefference: The mode_pipemode section of the pipeline can
               now contain subsections with changes to be applied to any other
               section in the pipeline configuration.
    2015-10-12 Marc Berthoud: PIPEMODE UPDATE: DRP now uses pipemode keyword with
               datakeys config file key to determine which pipeline mode should
               be run.
    2015-10-7 Marc Berthoud: Removed unused undo and resold 
    2014-7-31 Marc Berthoud, Added execute() to run pipeline from the
        command line.
    2012-9-27 Marc Berthoud, Added dynamic importing of modules for pipe steps
    2010-10-27 Marc Berthoud, Ver 0.1.1: Shift to PipeData 0.1.1 and upgrades:
        - Adapted to new requirement that pipe steps don't have config
        - Changed fileslist and reducedN to openfiles and reducedfiles
        - Changed to allow step name in configuration to be object name
        - Clean up call, error and undo mechanisms
    2008-11-21 Marc Berthoud, Ver 0.1: Wrote and Tested
"""



#! /usr/bin/env python

""" === DaRePype Run ===

    This tool facilitates setting up data reduction using the pipeline.
    The tool requires one (or several) pipeline run description file(s)
    which specify input files, configuration and output location.
    
    Usage:
      darepyperun pipe_run_description_file.txt
    OR
      darepyperun piperunfile1.txt piperunfile2.txt piperunfile3.txt  
    
    Pipe Run Description File:
    This file contains two sections: parameters and input file info.
    The following parameters can be used in the file
        pipeconf = folderpathname for pipeline config file !! REQUIRED !!
        pythonpath = list of paths to append to pythonpath
        pipemode = pipemode to use for reduction
        loglevel = level for logger (DEBUG / INFO (default) / WARN / ERROR)
        logfile = logfile(s) filepathname
        singlefile = T/F if True each file is run through a separate pipeline
                     instead of running all the files through at once (F is default)
        outputfolder = folder to write any results to (default is '.')
    The file can be specified in the following way
        inputfiles = 
        file1.fits
        file2.fits
        file10*.fits # same with wildcards
    Multipe pipeline configuration files and logfiles can be specified
    in the same manner as multiple input files are specified (but no
    "glob" style wildcards are allowed).
    The # character is used to specify comments, all text behind such
    a character is ignored.
    
    Arguments:
    * The only argument is the filename of the pipe run description file(s).
    * If no argument is given piperun looks for a pipe run description
      file called piperun.txt in the current folder.
    * If several pipeline description files are given, piperun will
      start each piperun with itself as a subprocess.
    * This tool has no other arguments, just call darepyperun directly
      (use 'python darepyperun -h' for details) to call the pipeline.
"""

### Basic Imports
import os 
import sys
import glob
import logging
import subprocess
import argparse

### Read run description
# Get file from arguments
parser = argparse.ArgumentParser(description = "Pipe Run")
parser.add_argument('prdfile', default='piperun.txt', type=str, nargs = '*',
                    help = 'pipe run description file name (default = piperun.txt)')
args = parser.parse_args()

### Run arguments
if len(args.prdfile) > 1:
    # Set up logger
    loglevel = 'INFO'
    logging.basicConfig(level=loglevel)
    log = logging.getLogger('piperun')
    # Get filepathname of piperun
    progfile = os.path.realpath(__file__)
    # Loop through piperuns
    for prunfile in args.prdfile:
        cmd = 'python %s %s' % (progfile, prunfile)
        log.info('Running Command: %s' % cmd)
        process = subprocess.Popen(cmd, shell=True )
        process.wait()
    # Exit
    exit()

### Process run description (get pythonpath)
# Read run description
rundesc = [l.strip() for l in open(args.prdfile[0])]
# Clear empty lines, strip comments, combine multi-line entries
rundict = {}
l1 = '' # full entry with several lines
rundesc.append('=') # to make sure last entry is added to rundict
for l in rundesc:
    # Search comment location and cut if found
    cloc = l.find('#')
    if cloc > -1: l = l[:cloc]
    # If line is new parameter, append old entry to 
    if '=' in l:
        eloc = l1.find('=')
        if eloc > -1:
            key = l1[:eloc].strip()
            val = l1[eloc+1:].strip()
            rundict[key]=val
        l1 = l
    # append line to previous entry
    else: l1 +='\n'+l

print(rundict)
# Look for pytonpath
if 'pythonpath' in rundict:
    sys.path.append(rundict['pythonpath'])

### Pipeline imports
from darepype.drp.pipeline import PipeLine
from darepype.drp.dataparent import DataParent

### Go through config options: Logfile Loglevel Pipemode
# Get outputfolder if available
if 'outputfolder' in rundict:
    outputfolder = rundict['outputfolder']
else:
    outputfolder = '.'
# Make outputfolder is required (needed before logging setup)
if not os.path.exists(outputfolder):
    os.mkdir(outputfolder)
# Set Logger: Get loglevel and set up
if 'loglevel' in rundict:
    loglevel = rundict['loglevel']
else:
    loglevel = 'INFO'
logging.basicConfig(level=loglevel)
# Set up logfiles if needed
if 'logfile' in rundict:
    logfiles = rundict['logfile']
    logfiles = logfiles.split('\n')
    logformat = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    for logf in logfiles:
        fhand = logging.FileHandler(logf)
        fhand.setFormatter(logging.Formatter(logformat))
        logging.getLogger().addHandler(fhand)
        print(logf)
# Setup log for piperun
log = logging.getLogger('piperun')
log.info('Setting Up: %s' % os.path.split(args.prdfile[0])[1])
# Get pipeconfig file
if 'pipeconf' in rundict:
    pipeconf = rundict['pipeconf']
else:
    log.error('Missing pipeconf in %s' % args.prdfile)
    raise ValueError('Missing pipeconf in %s' % args.prdfile)
# Load pipeconf, merge if there are multiple pipeconfs
pipeconf = pipeconf.split('\n')
log.debug('Loading Pipeconf=' + repr(pipeconf))
pipeconf = DataParent(config=pipeconf).config

# Get pipemode
if 'pipemode' in rundict:
    pipemode = rundict['pipemode']
else:
    pipemode = None
# Get singlefile
singlefile = False
if 'singlefile' in rundict:
    if rundict['singlefile'][0] in ['T','t']:
        singlefile = True

### Get input file list
# Option 1: Filenames (treat each with glob)
filelist = [] # list of input files to reduce
if 'inputfiles' in rundict:
    inputfiles = rundict['inputfiles'].split('\n')
    for infname in inputfiles:
        # Glob it
        fglob = glob.glob(infname.strip())
        fglob.sort()
        # If none found - Warning
        if len(fglob) == 0:
            log.warn("Glob: No files found for %s" % infname)
        # Add to filelist
        filelist += fglob
        
# Option 2: Folders (below inputfolder) and fileid

# No Files: Issue error and retunt
if len(filelist) < 1:
    log.error('No input files')
    raise ValueError('PipeRun: No input files')

### Copy Files and decompress if needed
filein = []
for fname in filelist:
    log.info("Copy file = " + os.path.split(fname)[1])
    if os.path.exists(fname.strip()):
        # Get new name (remove .bz2 if present)
        nameonly = os.path.split(fname)[1]
        if '.bz2' in fname[-4:]:
            nameonly = nameonly[:-4]
        newname = os.path.join(outputfolder,nameonly)
        # Copy the file
        if not os.path.exists(newname):
            os.system('cp ' + fname + ' ' + outputfolder)
            # Unzip if needed
            if '.bz2' in fname[-4:]:
                log.debug('Unzipping File ' + nameonly)
                os.system('bunzip2 '+ newname+'.bz2')
        # Add log / to filelist
        log.debug('Adding File ' + nameonly)
        filein.append(newname)

### Setup and Run pipeline
if singlefile:
    for f in filein:
        pipe = PipeLine(config=pipeconf)
        pipe(f, pipemode = pipemode, force=True)
else:
    pipe = PipeLine(config=pipeconf)
    pipe(filein, pipemode = pipemode, force=True)

### Final message
log.info('Finished piperun: %s' % os.path.split(args.prdfile[0])[1])
log.info("Reduced files:")
for fname in filein:
    log.info("   " + os.path.split(fname)[1])



""" PIPE DATA - Version 1.0.0

    This module defines the pipeline data parent object. The object
    has limited use to store, load and save data, it's child objects
    should be used instead.

    If the configuration file is properly populated, calling DataParent.load() or
    DataParent.loadhead() will return the proper DataChild object with the
    loaded file.

    The module also contains crucial variables for the entire pipeline,
    such as pipever and testconf.

"""

import numpy # numpy library
import os # os library
import logging # logging library
import time # time library
import configobj # config object library
import validate  # to check input config file is correct
import six # python 2 and 3 compatibility library
import re # regexp
import gc # garbage collect
#from __builtin__ import True # Not needed after python 2.X???
from fileinput import filename

class DataParent(object):
    """ HAWC Pipeline Data Object
        This object stores a header and the config file. Potential data is stored
        as a single image (numpy array).
    """
    # General variables: These are valid for all pipeline and all pipesteps
    pipever = '1.3.0beta3' # Pipeline version

    #testconf = 'config/pipeconf_master.txt' # Test configuration
    testconf = 'config/pipeconf_mgb.txt' # Test configuration

    # File Name Fit: Regexp expression that fits filenames
    filenamefit = '\A\Z' # Empty filename, only fits empty string

    def __init__(self, config=None):
        """ Constructor: Initialize data object and variables
        """
        # set up internal variables
        #     Remark: to preserve order of images no dictionary is used
        self.filename='' # Name of loaded file
        # Data Variable:
        self.data = None # Empty data array
        # Header: A dictionary. Lists for HISTORY and COMMENT entries.
        self.header = {}
        # set up logging and configuration
        self.log=logging.getLogger('pipe.data') # Logger
        self.config = None # Configuration
        self.setconfig(config) # Set the Configuration
        # Log message
        self.log.debug('Init: done')

    def __getattr__(self, name):
        """ Get Attribute function that allows to access these attributes:
            - filenamebegin: filename start. based on data:filenamebegin in
                             the configuration file, defaults to all before
                             .something.something.
            - filenameend: filename end. based on data:filenameend in the
                           configuration file, defaults to .something at the
                           end of the file.
            - filenum: file number. based on data:flilenum in the configuration
                       file. Returns None if there is no file number.
        """
        # return filenamebegin if it's requested
        if name == 'filenamebegin':
            (fpath,fname) = os.path.split(self.filename)
            # if filenamebegin is specified, use it
            try:
                filenamebegin = self.config['data']['filenamebegin']
                match = re.search(filenamebegin,fname)
            except (KeyError, TypeError):
                match = False
                filenamebegin = "None"
            if match :
                # found file name beginning -> return it
                return os.path.join(fpath,match.group())
            else:
                # assume filename format is name.filestep.fits or ....fits.gz
                msg = "Filename=%s doesn't match pattern=%s" % (fname, filenamebegin)
                self.log.warn(msg)
                extloc = fname.rfind('.f')
                #print extloc,fname
                if ( extloc < 0 or
                     fname[extloc:] not in ['.fts','.fits','.fits.gz'] ) :
                    self.log.warn('Filename has non-fits extension')
                    extloc = fname.rfind('.')
                    if extloc < 0: extloc = len(fname)
                    else: extloc += 1 # to add the '.'
                else: extloc += 1 # to add the '.'
                typeloc = fname[0:extloc-1].rfind('.')
                if typeloc < 0:
                    typeloc = extloc
                else:
                    typeloc += 1
                return os.path.join(fpath,fname[:typeloc])
        # return filenameend if it's requested
        if name == 'filenameend':
            (fpath,fname) = os.path.split(self.filename)
            # if filenameend is specified, use it
            try:
                filenameend = self.config['data']['filenameend']
                match = re.search(filenameend, fname)
            except (KeyError, TypeError):
                match = False
                filenameend = "None"
            if match :
                # found file name end -> return it
                return match.group()
            else:
                msg = "Filename=%s doesn't match pattern=%s" % (fname, filenameend)
                extloc = fname.rfind('.f')
                if ( extloc < 0 or
                     fname[extloc:] not in ['.fts','.fits','.fits.gz'] ) :
                    self.log.warn('Filename has non-fits extension')
                    extloc = fname.rfind('.')
                    if extloc < 0: extloc = len(fname)
                if extloc < 0:
                    return ''
                else:
                    return fname[extloc:]
        # return file number if it's requested
        if name=='filenum':
            (fpath,fname) = os.path.split(self.filename)
            # if filenum is specified in pipeconf, use it
            try:
                filenum = self.config['data']['filenum']
                match = re.search(filenum, fname)
            except (KeyError, TypeError):
                match = False
                filenum = None
            if match :
                # found file num -> return it
                for m in match.groups():
                    if m is not None:
                        return m
                return match.groups()
            else:
                return None
        # raise error if attribute is unknown
        msg="'%s' object has no attribute '%s'" % (type(self).__name__,name)
        raise AttributeError(msg)

    def setconfig(self,config):
        """ Sets configuration for the pipe data: The configuration
            object is returned. The config variable can be one of these
            - A ConfigObj object
            - A string to the filename of a valid config file
            - A list with strings to valid config files. In this case
              the first file is loaded, the others are merged with it.
              If all files have the same folder, only the first
              filepathname needs the full path.
        """
        if isinstance(config,configobj.ConfigObj):
            # if config is a ConfObj -> set it
            self.config=config
            self.log.debug('SetConfig: skipping configuration file validation')
            retmsg='received ConfigObj'
        elif isinstance(config,six.string_types):
            # if config is a string - check for file existence -> load it
            if os.path.isfile(config):
                try:
                    valpath = os.path.dirname(__file__)
                    valpath = os.path.join(valpath,'conf_validate.txt')
                    valpath = ''
                    self.config=configobj.ConfigObj(config,configspec=valpath)
                    #print self.config.configspec
                    #self.validateconfig() # taken out for now
                    # configuration validation couldn't keep up with
                    # new configuration parameters
                except configobj.ConfigObjError as error:
                    msg = 'Error while loading configuration file'
                    self.log.error('SetConfig: ' + msg)
                    raise error
                retmsg='loaded config file=' + config
            else:
                msg='<%s> is invalid file name for configuration' % config
                self.log.error('SetConfig: '+msg)
                raise IOError(msg)
        elif config is None:
            # no configuration -> No change
            # (( The following code would have been to use default config
            #    from the config validation file ))
            #self.config = None
            #path = os.path.dirname(__file__)
            #path = os.path.join(path,'conf_validate.txt')
            #self.config = configobj.ConfigObj(configspec=path)
            #self.config.validate(validate.Validator(),copy=True)
            #write out default
            #fp = open("pipe_test_conf.txt",'w')
            #self.config.write(fp)
            #fp.close()
            retmsg='no config given'
        else: # Assume it's a list-like object of config files
            # Get first object and check if it's really a list
            try:
                baseconf = config[0]
            except: # If it's not list-ish 
                # Invalid configuration - error
                self.log.error('SetConfig: Invalid configuration variable = %s' % repr(config))
                raise TypeError('Invalid configuration variable = %s' % repr(config))
            # Load first config
            self.setconfig(baseconf)
            basefolder = os.path.split(baseconf)[0]
            # Load other configs on top
            for conf in config[1:]:
                # Check if file exists
                if os.path.isfile(conf):
                    self.mergeconfig(conf)
                else:
                    cnf = os.path.join(basefolder, os.path.split(conf)[1])
                    if os.path.isfile(cnf):
                        self.mergeconfig(cnf)
                    else:
                        self.log.error('SetConfig: Invalid configuration file in list = %s' % conf)
                        raise ValueError('Invalid configuration file in list = %s' % conf)
            retmsg  = 'config list with %d files' % len(config)
        # Set environment variables from [envars]
        if not self.config is None:
            if 'envars' in self.config:
                for var in self.config['envars']:
                    os.environ[var] = str( self.config['envars'][var] )
        # Return
        self.log.debug('SetConfig: done ('+retmsg+')')
        return self.config

    def mergeconfig(self, newconfig):
        """ Merges a configuration object (or file) into the existing
            configuration. All values from the new configuration are used,
            overwriting old values if they are already in the old
            configuration.
        """
        # If there is no existing config: Just load the new config
        if not self.config:
            self.setconfig(newconfig)
            return
        # Store the existing config and load new config
        oldconf = self.config
        newconf = self.setconfig(newconfig)
        # Merge new to old config and store it
        oldconf.merge(newconf)
        self.config = oldconf
        
    def validateconfig(self):
        """ Test config against configspec and print errors if it doesn't
            conform.
            
            This code is currently not used

            Things that need to be fixed:
            * Config validation should not overwrite default value from
              paramlist in steps
            * If a keyword is missing in the [HEADER] section config validation
              shouldn't put it in.
        """

        errFlag = False # set to true if errors encountered
        results = self.config.validate(validate.Validator(),copy=True)
        extra   = configobj.get_extra_values(self.config) # keywords not in spec
        # List configuration keywords that failed validation
        if results != True:
            for (section_list, key,_) in configobj.flatten_errors(self.config,
                                                                  results):
                blah = ', '.join(section_list)
                if key is not None:
                    msg = "ValidateConfig: key '%s' in section '%s' failed"
                    self.log.error( msg % (key, blah))
                    errFlag = True
                else:
                    self.log.error("ValidateConfig: Section '%s' failed" %blah)
                    errFlag = True
        # Warn for keywords not found in validation
        for s,k in extra:
            if len(s) == 0:
                msg = 'ValidateConfig: Skippping unknown global keyword "%s"'
                self.log.warning(msg % k)
            else:
                msg = 'ValidateConfig: Skippping unknown keyword "%s" in section "%s"'
                self.log.warning( msg %(k,s[0]))
        if errFlag:
            raise validate.ValidateError
        self.log.debug('ValidateConfig: done')

    def getobject(self,objname):
        """ Looks for a module with the listed object. This function
            searches all package entries in config[general][steppacks] for
                package.objname
            If the module loads, it makes an objname object and returns it.
        """
        # get list of packages
        try:
            steppacks = self.config['general']['steppacks']
            if isinstance(steppacks,str): # ensure we have a list if only
                steppacks = [steppacks]    # 1 item, steppacks is str
        except KeyError as error:
            self.log.error('Setup: Missing steppacks item in configuration')
            raise error

        # Load object module
        stepmodule = None
        for pack in steppacks:
            if pack == '.': # look in local directory
                mod = objname.lower()
            else:
                mod = '%s.%s' % (pack,objname.lower())
            self.log.debug('looking for %s' % mod)
            # import the module
            try: 
                stepmodule = __import__(mod, globals(), locals(),
                                        [objname])
                self.log.debug('Pipe step %s found in %s' %
                              (objname,pack))
                break
            except (ImportError, ModuleNotFoundError) as msg:
                tmp = 'No module named %s' % objname.lower() # py2
                tmp1 = 'No module named \'%s.%s' % (pack,objname.lower()) # py3
                print('###',str(msg),tmp,tmp1)
                if str(msg).startswith(tmp) or str(msg).startswith(tmp1): # module not present in directory
                    self.log.debug('Pipe object %s not found in %s' %
                                   (objname, pack))
                else: # module tries to import a missing package
                    raise
            except:   # print out import errors not due to missing
                raise # modules (e.g., typos in code)
        # If not found -> Raise error
        if stepmodule == None:
            msg = 'Could not find object %s' % objname
            self.log.error('GetObject: ' + msg)
            raise ImportError(msg)
        # Make an object instance
        try:
            retobj = stepmodule.__dict__[objname]()
        except KeyError as error:
            msg = 'Pipe object %s' % objname
            msg+= ' not found in module %s' % mod
            self.log.error('GetObject: %s' % msg)
            raise error
        # Return the object
        return retobj

    def datamatch(self,filename=''):
        """ Returns an instance of the first data object in
            config['data']['dataobjects'] which matches the file under
            filename. This function returns an error if the file doesn't
            exist. The data is not loaded into the returned object.
        """
        # Get list of data objects
        try:
            dataobjects = self.config['data']['dataobjects']
            if isinstance(dataobjects,str): # ensure we have a list if only
                dataobjects = [dataobjects]    # 1 item, steppacks is str
        except KeyError as error:
            self.log.error('Setup: Missing dataobjects item in configuration')
            raise error
        # Loop through dataobjects
        found = False
        for dataname in dataobjects:
            # Get an object
            dataobj = self.getobject(dataname)
            dataobj.setconfig(self.config)
            # See if file fits
            if re.search(dataobj.filenamefit, filename) is not None:
                found = True
                break
        if not found:
            msg = 'DataMatch: No matching data object in %s for filename %s' % (repr(dataobjects), filename)
            self.log.error(msg)
            raise ValueError(msg)
        else:
            msg = 'DataMatch: Found matching data object %s for filename %s' % (dataname, filename)
            self.log.debug(msg)
        # Return the object
        return dataobj

    def loadhead(self,filename=''):
        """ Searches for a matching dataobject for the file given, then
            calls the loadhead() function of that object on the file.
            The new dataobject is returned.
            !! Only DataParent.loadhead returns such an object,     !!
            !! no return is expected from loadhead of child objects !!
        """
        # set self.filename and filename
        if len(filename) > 0:
            self.filename=filename
        else:
            filename=self.filename
        # Check for file existance
        if not os.path.exists(filename):
            msg = 'LoadHead: %s is invalid filename' % filename
            self.log.error(msg)
            raise ValueError(msg)
        # Create a data object for the file
        dataobj = self.datamatch(filename)
        # Load the header for the parent object
        dataobj.loadhead(filename)
        # Return the parent object
        return dataobj

    def load(self,filename=''):
        """ Searches for a matching dataobject for the file given, then
            calls the load() function of that object on the file.
            The new dataobject is returned.
            !! Only DataParent.load returns such an object,     !!
            !! no return is expected from load of child objects !!
        """
        # set self.filename and filename
        if len(filename) > 0:
            self.filename=filename
        else:
            filename=self.filename
        # Check for file existance
        if not os.path.exists(filename):
            msg = 'Load: %s is invalid filename' % filename
            self.log.error(msg)
            raise ValueError(msg)
        # Create a data object for the file
        dataobj = self.datamatch(filename)
        # Load the header for the parent object
        dataobj.load(filename)
        # Return the parent object
        return dataobj

    def save(self,filename = ''):
        """ Save the data in the object to the specified file. Existing files are
            overwritten. This issues a warning (dataparent should not be used for
            storing data) then searches for a matching dataobject to the given
            filename. Data and header are copied to that dataobject and the data
            is saved.
            - filename: a string with the file name to save to
        """
        # Issue warning
        self.log.warn('DataParent.save should not be used as DataParent is not ' +
                      'intended to store data.')
        # get file name
        if len(filename) > 0:
            self.filename = filename
        else:
            filename = self.filename
        # Create data object for the file
        dataobj = self.datamatch(filename)
        # Set filename
        dataobj.filename = filename
        # Copy header
        for key in self.header.keys():
            if key != 'COMMENT' and key != 'HISTORY':
                dataobj.setheadval(key,self.header[key])
        # Copy Comments
        if 'COMMENT' in self.header:
            for com in self.header['COMMENT']:
                dataobj.setheadval('COMMENT',com)
        # Copy History
        if 'HISTORY' in self.header:
            for hist in self.header['HISTORY']:
                dataobj.setheadval('HISTORY',hist)
        # Copy data
        dataobj.data = self.data
        # Save
        dataobj.save()

    def copy(self):
        """ Returns a copy of self
        """
        out=DataParent(config=self.config) # create new object
        # copy filename and header
        out.filename = self.filename
        out.header = self.header.copy()
        # Copy data - backup if no copy() available
        try:
            out.data = out.data.copy()
        except:
            out.data = out.data
        # return message and new object
        self.log.debug('Copy: done')
        return out

    def mergehead(self, other):
        """ Merges the header of another data object to the existing header.
            Most of the header of the new data is ignored, PARENT,
            HISTORY keywords are copied.
        """
        # get selfhist and otherhist lists
        if 'HISTORY' in self.header:
            selfhist = self.header['HISTORY']
        else: selfhist = []
        if 'HISTORY' in other.header:
            otherhist = other.header['HISTORY']
        else: otherhist = []
        # add history keywords (no duplicates)
        selfhist += [hist for hist in otherhist if not hist in selfhist]
        # if there is something add write back to header
        if len(selfhist) : self.header['HISTORY'] = selfhist
        # get selfcomm and othercomm lists
        if 'COMMENT' in self.header:
            selfcomm = self.header['COMMENT']
        else: selfcomm = []
        if 'COMMENT' in other.header:
            othercomm = other.header['COMMENT']
        else: othercomm = []
        # add comment keywords (no duplicates)
        selfcomm += [comm for comm in othercomm if not comm in selfcomm]
        # if there is something add write back to header
        if len(selfcomm) : self.header['COMMENT'] = selfcomm

        # Go through keywords listed in headmerge: assume self is first
        headmerge = self.config['headmerge']
        for key in headmerge.keys():
            if key in self.header and key in other.header:
                selfval = self.header[key]
                otherval = other.header[key]
                operation = headmerge[key].upper()
                if operation == 'LAST': selfval = otherval
                elif operation == 'MIN': selfval = min(selfval,otherval)
                elif operation == 'MAX': selfval = max(selfval,otherval)
                elif operation == 'SUM': selfval += otherval
                elif operation == 'OR': selfval = selfval | otherval
                elif operation == 'AND': selfval = selfval & otherval
                elif operation == 'CONCATENATE':
                    if ',' in selfval:
                        vlist = selfval.split(',')
                    else:
                        vlist = [selfval]
                    if otherval not in vlist:
                        vlist.append(otherval)
                        selfval = ','.join(sorted(vlist))
                elif operation == 'DEFAULT':
                    if type(selfval) is str:
                        selfval = 'UNKNOWN'
                    elif type(selfval) is int or type(selfval) is long:
                        selfval = -9999
                    elif type(selfval) is float:
                        selfval = -9999.0
                self.header[key] = selfval
        self.log.debug('MergeHead: done')

    def getheadval(self, key, errmsg=True):
        """ Get Header Value: Returns the value of the requested key from
            the header. If the key is present in the [Header] section
            of the configuration that value is returned instead, the following
            entries are possible in the configuration file:
              * KEY = VALUE * VALUE is returned, the system checks if value is an
                              int or a float, else a string is returned.
              * KEY = NEWKEY * The VALUE under header[NEWKEY] is returned.
              * KEY = ?_ALTKEY * If the keyword KEY is present, header[KEY] is
                                 returned, else header[ALTKEY] is returned.
            If the key can not be found an KeyError is produced and a warning is
            issued (unless key is present in the [Header] section of the
            configuration).

            errmsg: Flag indicating if a log error message should be
                    issued if the keyword is not found. This can be disabled
                    (set it to False) if getheadval is used to probe a dataset.
        """
        val = None
        inkey = key # retain key which was input in case key changes
        # Look in the config
        try:
            # get the value
            val = self.config['header'][key]
            # Check if it's optional header replacement i.e. starts with '?_'
            if val[:2] in ['?_', '? ', '?-']:
                # if key is not in the header -> use key name under value instead
                if not key in self.header:
                    key = val[2:].upper()
                val = None
            # Check if it's a Header replacement (but not T/F)
            elif val[0].isalpha() and val[:2] not in ['T ', 'F '] and val not in ['T', 'F']:
                self.log.info('Getheadval: Using %s value for %s' %
                              (val.upper(), key ))
                key = val.upper()
                val = None
            # Else: read value
            else:
                # Try as T / F
                found = True
                if val == 'T' or val[:2] == 'T ':
                    val = True
                elif val == 'F' or val[:2] == 'F ':
                    val = False
                else:
                    found = False
                # Try as int
                if not found:
                    try:
                        val = int(val)
                        found = True
                    except:
                        pass
                # Try as float
                if not found:
                    try:
                        val = float(val)
                        found = True
                    except:
                        pass
                # If not found - just leave value as string
                # update value in header
                self.setheadval(key, val)
        except KeyError :
            # if key is not in config - continue
            pass
        except TypeError :
            # if config is not yet loaded - return error
            self.log.warn('GetHeadVal: Missing Configuration')
            # The following line is commented out such that pipedata can
            #     still be used without configuration
            #raise RuntimeError('Missing Pipe Configuration')
        # Look in the header
        if val is None:
            # get value from header
            try:
                val = self.header[key]
            except KeyError:
                # if keyword is not found
                msg='Missing %s keyword in header' % key
                if errmsg:
                    self.log.error('GetHeadVal: %s' % msg)
                raise KeyError(msg)

        # Debug message (only first line if multi-line value)
        if '\n' in str(val):
            self.log.debug('GetHeadval: done (%s=%s ...)' % (inkey, str(val[0])))
        else:
            self.log.debug('GetHeadVal: done (%s=%s)' % (inkey,str(val)))
        return val

    def setheadval(self, key, value, comment = ''):
        """ Sets a FITS keyword in the header
        """
        # If key==HISTORY or COMMENT: add to history list
        if key=='HISTORY' or key=='COMMENT':
            if key in self.header:
                self.header[key].append(value)
            else:
                self.header[key] = [value,]
        # Else: add as normal keyword
        else:
            self.header[key] = value
            if len(comment) > 0:
                self.setheadval('COMMENT','%s, %s' % (key,comment))
        self.log.debug('SetHeadVal: done')

    def delheadval(self, key):
        """ Delete one or more FITS keyword in specified header, which defaults
            to the first header.

            WARNING: This will delete all COMMENT or HISTORY entries
        """
        # If key is a list, remove all entries
        if isinstance(key,(list,tuple)):
            for k in key:
                self.delheadval(k)
        # Else if it's a string delete the key - this may rise KeyError
        else:
            del(self.header[key])
        self.log.debug('DelHeadVal: done')

    def test(self):
        """ Test Pipe Data Object: Runs a set of tests on a pipedata object.
            This function needs to start in the folder contining the test
            files.
        """
        self.log.info("Testing Data Object")
        ### Setup
        # test loading config
        #self.setconfig(1) # should raise TypeError
        self.setconfig(None)
        conf = self.testconf
        self.setconfig(conf)
        conf=configobj.ConfigObj(conf)
        self.setconfig(conf)
        ### Loading and Saving DAta
        # self.load(1) # should create TypeError
        # self.load('nothere.fits') # should create IOError
        # self.load('makebasic.pro') # should create IndexError
        testpath = self.config['testing']['testpath']
        print(testpath+'/makedata/testimgtabimg.fits')
        self.load(testpath+'/makedata/testimgtabimg.fits')
        self.log.debug('Images = ' + repr(self.imgnames))
        self.save(testpath+'/testsave.fits')
        print(self.filenamebegin, self.filenameend)
        #self.save(1) # creates TypeError
        ### Copy, Merge and Add Data
        other=self.copy()
        other.mergedata(self)
        other.load(testpath+'/makedata/testimgtabimg.fits')
        other.imgdata[1]=numpy.zeros((10,20))
        #other.table = numpy.rec.array([(1)], names='nr', formats='i2')
        self.mergedata(other)
        self.adddata({'First Image':numpy.zeros((10,10))},['Delta',4])
        for i in range(len(self.imgdata)):
            s='%s %s' % (self.imgnames[i],self.imgdata[i].shape)
            self.log.debug(s)
        self.log.debug(repr(self.table))
        ### Images / Table / Header
        # Images
        img=self.imageget('SECOND IMAGE')
        self.imageset(numpy.zeros((5,5,5)),'First Image Copy')
        self.imagedel('SECOND IMAGE')
        self.imageset(img,'Second Image',index=-1)
        other.imgnames[0]='WILDCARD'
        self.copydata(other,'WILDCARD')
        # Image variable
        self.image=numpy.zeros((5,5))
        self.image[0,:]=range(5)
        # Table
        i=len(self.table)
        #print(self.table,self.table.dtype,self.table.shape)
        self.tableaddcol('index',numpy.arange(i)-2)
        self.tableaddrow(['Zeta',0,100])
        self.tabledelcol('NR')
        self.tabledelrow(3)
        #print(self.table,self.table.dtype,self.table.shape)
        self.table['index']=numpy.arange(len(self.table))+2
        self.table[0]=('first',100)
        self.tableaddcol('arr', numpy.ones((7,2)))
        self.log.debug(repr(self.table))
        self.log.debug(repr(self.table.dtype))
        self.log.debug('TabMrg='+repr(self.tablemergerows(self.table)))
        # Second table
        tab2 = self.tableget()
        self.tableset(tab2,'Second Table')
        print(self.tableget('Second Table').dtype,self.tabdata[1])
        self.tabledelcol('arr','Second Table')
        self.tabledelrow(6,'Second Table')
        self.tabledelrow(5,'Second Table')
        self.tableaddrow(['velo',19], 'Second Table')
        self.tableaddcol('NewNr', range(6), 'Second Table')
        print(self.tableget('Second Table').dtype,self.tabdata[1])
        # Delete first table and save
        self.tabledel('TABLE')
        self.save(testpath+'/testsave.fits')
        # Header
        self.header['Reduced'] = 0
        self.setheadval('TST','Test Value','') #,'Second Table')
        print(self.header)
        self.getheadval('TST') #,'Second Table')
        self.setheadval('TST',15)
        print(self.header)
        self.setheader(self.getheader())
        # Additional Load and Save Tests
        if os.path.exists(testpath+'/makedata/outtest.fits'):
            os.remove(testpath+'/makedata/outtest.fits')
        self.save(testpath+'/makedata/outtest.fits')
        #self.save('pipedata/outtest.fits')
        #self.load('pipedata/outtest.fits')
        #self.load('sample.chop_nod.dog.raw.fits')
        #self.load('sample.noise.fft.raw.fits')
        self.log.info("Test Finished")
        print('Done')
        return 0

if __name__ == '__main__':
    """ Main Function for testing:
        Changes to the folder with the test files and runs pipedata.test()
    """
    logging.basicConfig(level=logging.DEBUG)
    data=DataParent()
    data.test()

""" === History ===
    2016-9-26 Marc Berthoud: Split PipeData (fits only) into
        DataParent and DataFits
        Main Changes:
        - init no longer accepts a file name (since it can't return
          the correct object.
        - each object has a filenamefit variable (is fitted with regexp)
    2015-9-25 Marc Berthoud: Edited save() to update keywords for
        FILENAME and DATE
    2014-7-31 Marc Berthoud: Added output_verify=fix to write()
        Fixed tablemergerows to work for FITS_rec
        Now use numpy.insert for tableaddrow, numpy.delete in tabledelrow
    2014-2-17 Marc Berthoud: Fix error with load if primary header empty
    2012-11-15 Marc Berthoud: Loadhead now loads from other HDUs
    2012-9-27 Marc Berthoud: Add functionality for header keyword replacement
    2012-9-13 Marc Berthoud: Allow files to have empty HDUs
        i.e. image.shape=(0,)
    2012-5-17 Marc Berthoud: Added storate and management of headers for all
    2012-2-17 Marc Berthoud: Added multiple tables option
        Many new commands (tableindex/get/set/del) and upgrades to table
        commands.
        COMMENT: adddata now only affects first table (may have to change)
    2011-2-4 Marc Berthoud: Added filenamebase to __getattr__
    2010-12-20 Marc Berthoud: Made sure that all image names would always be
        uppercase.
    2010-12-3 Marc Berthoud: Added code to tableaddcol such that arrays
        can be stored in tables.
    2010-10-1  Marc Berthoud, Ver0.1.1: New structure for images:
        - Now there can be multiple images
        - The size of the primary image doesn't have to agree with the
          table size
    2009-11-17 Marc Berthoud, Added remap to sky index
    2008-10-28 Marc Berthoud, Ver0.1: Wrote and tested
"""



#!/usr/bin/env python
""" PIPE STEP LOG SOCKET - Version 1.0.0

    This module runs a background socket connection that accepts log messages to
    be added to the pipeline log. The socket should be parent of all pipeline
    steps which use that socket. To start the socket connection logsocketstart
    has be be called in the setup() function of these steps, repeated calls to
    logsocketstart are expected and ignored. The socket connection runs in a
    separate thread. Configuration settings for logsocket are expected in the
    [logsocket] section of the configuration file.

    This object is a child of StepParent so that self.log is available and the
    -t option can be used for testing.

    @author: berthoud
"""

import os # os library
import socket # socket library
import _thread # thread library
import time # time libary
import logging # logging object library
from ..drp.stepparent import StepParent

class StepLogSocket(StepParent):
    """ Pipeline LogSocket Object
        Should be parent object of pipesteps that require
        the socket.
    """
    socketactive = [False,]
    activeloghost = [None,]
    activelogport = [None,]

    def logsocketstart(self):
        """ Opens the log socket if it is not yet active
        """
        # Check if socket is active
        if not self.socketactive[0]:
            # Set up socket
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            try:
                conf = self.config['logsocket']
            except:
                self.log.warn('Missing [logsocket] configuration ' +
                              '- using default setup: localhost:50747')
                conf={'host':'localhost','port':50747}
            # Try specified port + up to 20
            cport = int(conf['port'])
            err = False
            for i in range(20):
                try:
                    self.log.debug('Attempting port ' + str(cport))
                    sock.bind((conf['host'],cport))
                    # Start listening
                    sock.listen(5)
                    # Start thread
                    thread.start_new_thread(self.logsocketrun,(sock,))
                    err = False
                    break
                except Exception:
                    err = True
                    cport += 1
                    continue
            # If still failed, then raise error
            if err: raise

            # Save host and port
            self.activeloghost[0] = socket.gethostbyname(conf['host'])
            self.activelogport[0] = cport

            self.log.info('Started Logging Socket on port %d' % cport)
            self.socketactive[0]=True
        else:
            self.log.debug('Multiple calls to LogSocketStart - socket already running')

        self.loghost = self.activeloghost[0]
        self.logport = self.activelogport[0]

    def logsocketrun(self, sock):
        """ Log socket subprocess loop
        """
        try:
            conf = self.config['logsocket']
        except:
            self.log.warn('Missing [logsocket] configuration ' +
                          '- using default backup logger: pipe.logserver')
            conf={'deflogger':'pipe.logserver'}
        while 1:
            # Get new connection
            conn, addr = sock.accept()
            #print('Conected with %s at address %s' % (addr[0],str(addr[1])))
            # Get the message
            reply = conn.recv(1024)
            #print('  Got message: %s' % reply)
            # Close the connection
            conn.close()
            #print('  Connection Closed')
            ### Make log message
            #print('  Making log message:')
            # Split incoming message
            split = reply.split('\t')
            if len(split) < 2:
                lvl = 'INFO'
                lgr = conf['deflogger']
                msg = split[0]
            elif len(split) < 3:
                lvl = split[0]
                lgr = conf['deflogger']
                msg = split[1]
            else:
                lvl = split[0]
                lgr = split[1]
                msg = '\t'.join(split[2:])
            # Get logging level
            log=logging.getLogger(lgr)
            if lvl.upper() in ['DEBUG','DEBUGGING']: log.debug(msg)
            elif lvl.upper() in ['INFO','INFORMATION']: log.info(msg)
            elif lvl.upper() in ['WARN','WARNING']: log.warn(msg)
            elif lvl.upper() in ['ERR','ERROR']: log.error(msg)
            elif lvl.upper() in ['CRIT','CRITICAL']: log.critical(msg)
            else: log.info(msg)
            #if lvl.upper() in ['DEBUG','DEBUGGING']: lvl=logging.DEBUG
            #elif lvl.upper() in ['INFO','INFORMATION']: lvl=logging.INFO
            #elif lvl.upper() in ['WARN','WARNING']: lvl=logging.WARN
            #elif lvl.upper() in ['ERR','ERROR']: lvl=logging.ERROR
            #elif lvl.upper() in ['CRIT','CRITICAL']: lvl=logging.CRITICAL
            #else: lvl=logging.INFO
            #rec=logging.LogRecord(lgr,lvl,'','',msg,{},{})
            #log.handle(rec)

    def test(self):
        """ Test Pipe Step Parent Object:
            Runs a set of basic tests on the object
        """
        # log message
        self.log.info('Testing pipe step %s' %self.name)
        # set up logger
        self.logsocketstart()
        # wait
        time.sleep(1)
        # send log message via python socket
        self.log.debug('Python Socket Test:')
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect(('localhost', 50747))
        message = 'error\tpipe.logtest.client\tTest Message'
        s.sendall(message)
        s.close()
        # send log message via c socket
        self.log.debug('C-Program Test')
        os.system("/Users/berthoud/tech/tcpip/simpleclient $'warn\tpipe.logtest.C\tWarn Message'")
        os.system("/Users/berthoud/tech/tcpip/simpleclient $'info\tpipe.logtest.C\tInfo Message'")
        os.system("/Users/berthoud/tech/tcpip/simpleclient $'debug\tpipe.logtest.C\tDebug Message'")
        # wait
        time.sleep(1)
        # log message
        self.log.info('Testing pipe step %s - Done' %self.name)

if __name__ == '__main__':
    """ Main function to run the pipe step from command line on a file.
        Command:
          python stepparent.py input.fits -arg1 -arg2 . . .
        Standard arguments:
          --config=ConfigFilePathName.txt : name of the configuration file
          -t, --test : runs the functionality test i.e. pipestep.test()
          --loglevel=LEVEL : configures the logging output for a particular level
          -h, --help : Returns a list of
    """
    StepLogSocket().execute()

""" === History ===
    2015-4-5 Marc Berthoud: Ver 0.1.0: Written and tested
"""



""" PIPE DATA FITS - Version 1.0.0

    This module defines the pipeline FITS data object. The object
    is used by the pipeline to store raw, intermediate and reduced data.

"""

import numpy # numpy library
import os # os library
import logging # logging library
import time # time library
import configobj # config object library
import validate  # to check input config file is correct
import re # regexp
import gc # garbage collect
from astropy.io import fits # fits libary
from .dataparent import DataParent # Pipe Data Parent
#from __builtin__ import True # Not needed after python 2.X???

class DataFits(DataParent):
    """ Pipeline Data FITS Object
        The data is stored as a list of (multi-dimensional) images and a
        similar list of tables. Fits  file name and all headers are
        stored as well.
    """

    # File Name Fit: Regexp expression that fits valid filenames
    filenamefit = r'\.(fits|fts|fit)\Z' # will fit .fits or .fts at end of name

    def __init__(self, filename='', config=None):
        """ Constructor: Initialize data object and variables
        """
        # set up internal variables
        #     Remark: to preserve order of images no dictionary is used
        self.filename=filename # Name of loaded file
        # Image Variables:
        #   At all times len(imgdata) == len(imgnames) == len(imgheads)
        #   - If only loadhead() was called or first HDU has no data:
        #     imghead[0]=header, imgdata[0]=None, imgnames[0]='Primary Header'
        self.imgdata=[] # List with image data - data is in array of int/float
        self.imgnames=[] # List with image names (all uppercase)
        self.imgheads=[] # List with image headers
        # Table Variables:
        #   At all times len(tabdata) == len(tabnames) == len(tabheads)
        self.tabdata=[] # List with tables each item is a record array
        self.tabnames=[] # List with table names (all uppercase)
        self.tabheads=[] # List with table headers
        # set up logging and configuration
        self.log=logging.getLogger('pipe.datafits') # Logger
        self.config = None # Configuration
        self.setconfig(config) # Set the Configuration
        # Count to decrease number of frequent log messages
        self.logcount=0
        # if file exists, load it
        if os.path.exists(filename):
            self.load(filename)
        elif filename != '': # user specified a non-existant filename
            msg = 'No such file %s' %filename
            self.log.error(msg)
            raise IOError(msg)
        self.log.debug('Init: done')

    def __getattr__(self, name):
        """ Get Attribute function that allows to access these attributes:
            - header: header of primary HDU
            - image: the first image
            - table: the first table
            - filenamebase: filename start - depreciated
            From DataParent.__getattr__:
            - filenamebegin: filename start
            - filenameend: filename end
        """
        # return data
        if name == 'data':
            # return image if available
            if len(self.imgdata) > 0 and self.imgdata[0] != None:
                return self.imgdata[0]
            # else return a table
            elif len(self.tabdata) > 0:
                return self.tabdata[0]
            else:
                msg="'%s' object has no data" % (type(self).__name__)
                self.log.warn(msg)
                return None
        # return image if it's requested - raise error if no images present
        if name == 'image':
            if len(self.imgdata) == 0:
                msg="'%s' object has no image data" % (type(self).__name__)
                self.log.warn(msg)
                return None
            elif self.imgdata[0] is None:
                msg="'%s' object has no image data" % (type(self).__name__)
                self.log.warn(msg)
            return self.imgdata[0]
        # return table if it's requested - raise error if no tables present
        if name == 'table':
            if len(self.tabdata) == 0:
                msg="'%s' object has no table data" % (type(self).__name__)
                self.log.warn(msg)
                return None
            return self.tabdata[0]
        # return header if it's requested - raise error if no images present
        if name == 'header':
            if len(self.imgheads) == 0:
                msg="'%s' object has no header data" % (type(self).__name__)
                self.log.warn(msg)
                return None
            return self.imgheads[0]
        # return filenamebase if it's requested
        if name == 'filenamebase':
            self.log.warn('Filenamebase is depreciated - use filenamebegin')
            return self.filenamebegin
        # run parent function (filenamebegin and filenameend)
        return super(DataFits, self).__getattr__(name)
        # raise error if attribute is unknown
        msg="'%s' object has no attribute '%s'" % (type(self).__name__,name)
        raise AttributeError(msg)

    def __setattr__(self, name, value):
        """ Set Attribute function that allows to access the first
            image as PipeData.image
        """
        # set the data to image
        if name == 'data':
            # if it's a table put it in table
            if issubclass(value.__class__, numpy.recarray):
                self.table = value
            # else assume it's an image
            else:
                self.image = value
        # set the image if it's requested
        if name == 'image':
            if len(self.imgdata) > 0:
                self.imgdata[0] = value
            else:
                self.imgdata=[value]
                self.imgnames=['PRIMARY IMAGE']
                self.imgheads=[fits.Header()]
        # set the table if it's requested
        elif name == 'table':
            if len(self.tabdata) > 0:
                self.tabdata[0] = value
            else:
                self.tabdata=[value]
                self.tabnames=['PRIMARY TABLE']
                self.tabheads=[fits.Header()]
        # set the header if it's requested
        elif name == 'header':
            if len(self.imgdata) > 0:
                self.imgheads[0] = value
            else:
                self.imgdata=[None]
                self.imgnames=['PRIMARY HEADER']
                self.imgheads=[value]
        # else pass the command to the parent function
        else:
            object.__setattr__(self, name, value)

    def loadhead(self,filename='', dataname=''):
        """ Loads and returns the primary header of the FITS file given.
            This also checks for file existence and type. The pipedata object
            is filled with only that header.
            - filename: a string with the name of the file to load
              (if omitted, self.filename is used)
            - dataname specified the EXTNAME value of the header to be loaded
              If such a header is not found, or dataname=='', the first header
              is loaded. This option should be used if the main file
              information is not in the primary header.
        """
        ### check for file existence, type and get primary header
        # set self.filename and filename
        if len(filename) > 0:
            self.filename=filename
        else:
            filename=self.filename
        # read fits header, checks for existing valid fits file
        try:
            hdus = fits.open(filename)
        except IOError as error:
            # IOError: 2 No such file or directory (missing file)
            self.log.error('LoadHead: no such file or head read error, '
                           + filename)
            raise error
        except IndexError as error:
            # IndexError: list index out of range (invalid format for fits)
            self.log.error('LoadHead: '+filename+' is not a valid FITS file')
            raise error
        except TypeError as error:
            # TypeError: invalid filename (not a string)
            self.log.error('LoadHead: filename is invalid type')
            raise error
        # No dataname -> return primary header
        if dataname == '':
            header = hdus[0].header
        # Look for correct dataname
        else:
            hdui = 0
            found = 0
            while not found:
                if 'EXTNAME' in hdus[hdui].header.keys():
                    if hdus[hdui].header['EXTNAME'].strip() == dataname:
                        found = 1
                        header = hdus[hdui].header
                hdui += 1
            if not found:
                msg = "loadhead: HDU with EXTNAME=%s not found" % dataname
                self.log.error(msg)
                raise ValueError(msg)
        # Fill in the header, if necessary fill in data
        self.imgheads = [header]
        self.imgdata = [None]
        try:
            self.imgnames = [header['EXTNAME'].upper()]
        except:
            self.imgnames = ['PRIMARY HEADER']
        # Fill the filename
        self.filename = filename
        self.log.debug('LoadHead: done')
        hdus.close()

    def load(self,filename=''):
        """ Load a file into the data object. The file can be a raw HAWC
            file or a file saved by pipedata.
            - filename: a string with the name of the file to load
              (if omitted, self.filename is used)
            - raises various errors if file can not be read
        """
        ### Clear file data
        self.imgdata, self.imgheads, self.imgnames = [], [], []
        self.tabdata, self.tabheads, self.tabnames = [], [], []
        ### Get filename and file checks
        # set self.filename and filename
        if len(filename) > 0:
            self.filename=filename
        else:
            filename=self.filename
        # check for file existence, type and get primary header
        #   fills first entry into imgdata, imgnames
        self.loadhead(filename)
        ### Read File
        hdus=fits.open(filename,memmap=False)
        ### Collect images / Load them
        # Search for ImageHDUs (does not include PrimaryHDU)
        imgind=[i for i in range(len(hdus))
                if isinstance(hdus[i], fits.ImageHDU) ]
        imgn = len(imgind) # store number of images
        # get naxis and naxis1 from primary header, check if keywords exist
        try:
            naxis=self.getheadval('NAXIS')
            if naxis > 0:
                naxis1=self.getheadval('NAXIS1',errmsg=False)
            else:
                naxis1=0
        except KeyError:
            # KeyError: keyword not found (no keywords available)
            self.log.warn('Load: missing naxis keywords in fits file '
                           + filename)
        # Check if file has no image (i.e. if naxis==1 and naxis1==0)
        # -> No change, primary image stays as it was from loadhead()
        if naxis * naxis1 == 0:
            self.log.info('Load: No image data in first HDU')
        # else: Load first HDU data if there is image data in it
        else:
            imgn += 1
            # Load data
            self.imgdata[0] = hdus[0].data
            # Set image name
            if 'EXTNAME' in self.imgheads[0]:
                self.imgnames[0] = self.imgheads[0]['EXTNAME'].upper()
            else:
                self.imgnames[0] = 'PRIMARY IMAGE'
        # Message with number of images
        if imgn == 0:
            self.log.debug('Load: No image in file %s' % filename)
        elif imgn == 1:
            self.log.debug('Load: 1 image in file %s' % filename)
        else:
            self.log.debug('Load: %d images in file %s' %(imgn, filename))
        # Load subsequent images
        for ind in imgind:
            # get name
            if 'EXTNAME' in hdus[ind].header.keys():
                self.imgnames.append(hdus[ind].header['EXTNAME'].upper())
            else:
                self.imgnames.append('SECONDARY IMAGE %d' % ind)
            # get data and header
            self.imgdata.append(hdus[ind].data)
            self.imgheads.append(hdus[ind].header)
        ### Look for Tables / Read them
        # Search for BinTableHDUs
        tabind=[i for i in range(len(hdus))
                if isinstance(hdus[i], fits.BinTableHDU) ]
        tabn = len(tabind)
        # Messages on number of tables
        if tabn == 0:
            self.log.debug('Load: No table in file %s' % filename)
        elif tabn == 1:
            self.log.debug('Load: %d table in file %s' %(tabn, filename))
        else:
            self.log.debug('Load: %d tables in file %s' % (tabn, filename))
        # Load all tables
        for ind in tabind:
            # if table is empty -> warning and skip it
            try:
                if hdus[ind].data is None:
                    msg = 'Load: Table in HDU number %d has no data' % ind
                    msg += ' -> Ignoring this HDU'
                    self.log.warn(msg)
                    continue
            except:
                msg = 'Load: Problem loading table in HDU number %d' % ind
                msg += ' -> Ignoring this HDU'
                self.log.warn(msg)
                continue

            # get name
            if 'EXTNAME' in hdus[ind].header.keys():
                self.tabnames.append(hdus[ind].header['EXTNAME'].upper())
            else:
                if ind > 0:
                    self.tabnames.append('SECONDARY TABLE')
                else:
                    self.tabnames.append('PRIMARY TABLE')
            # get data, header
            self.tabdata.append(numpy.rec.array(hdus[ind].data))
            self.tabheads.append(hdus[ind].header)
        # Close the file
        hdus.close()
        gc.collect()
        ### Message
        self.log.debug('Load: loaded fits file')

    def save(self,filename = None):
        """ Save the data in the object to the specified file. Existing files are
            overwritten.
            - filename: a string with the file name to save to
        """
        # get file name
        if filename == None:
            filename = self.filename
        # update pipeline keywords
        self.setheadval('PIPEVERS', DataParent.pipever.replace('.','_'),
                        'Pipeline Version')
        self.setheadval('FILENAME',os.path.split(filename)[-1])
        self.setheadval('DATE',time.strftime('%Y-%m-%dT%H:%M:%S'))
        ### save file
        # make the data primary HDU -> List
        hdul = [] # hdu list
        for i in range(len(self.imgnames)):
            if i == 0:
                hdui = fits.PrimaryHDU(self.imgdata[i],self.imgheads[i])
            else:
                hdui = fits.ImageHDU(self.imgdata[i],self.imgheads[i])
            hdui.header['EXTNAME'] = (self.imgnames[i].upper(),'ID of the HDU')
            hdul.append(hdui)
        if len(hdul) == 0: # no IMAGE HDUs added
            hdul.append(fits.PrimaryHDU(None,self.imgheads[0]))
        # make hdus for tables
        for i in range(len(self.tabnames)):
            hdut = fits.BinTableHDU(self.tabdata[i],self.tabheads[i])
            hdut.header['EXTNAME'] = (self.tabnames[i].upper(),'ID of the HDU')
            hdul.append(hdut)

        # make an HDU list
        hdulist=fits.HDUList(hdul)
        # save the file (produce errors if not successful)
        try:
            hdulist.writeto(filename, output_verify='fix', overwrite=True)
        except TypeError as error:
            self.log.error('Save: filename is invalid type')
            raise error
        except IOError as error:
            self.log.error('Save: Failed to write fits file to '+filename)
            raise error
        self.log.debug('Save: wrote FITS file %s' % filename)

    def copy(self):
        """ Returns a copy of self
        """
        out=DataFits(config=self.config) # create new object
        # copy all images
        out.imgnames=self.imgnames[:]
        out.imgdata=[]
        out.imgheads=[]
        for imgi in range(len(self.imgdata)):
            if self.imgdata[imgi] is not None:
                out.imgdata.append(self.imgdata[imgi].copy())
            else:
                out.imgdata.append(None)
            out.imgheads.append(self.imgheads[imgi].copy())
        # copy tables
        out.tabnames=self.tabnames[:]
        out.tabdata=[]
        out.tabheads=[]
        for tabi in range(len(self.tabdata)):
            out.tabdata.append(self.tabdata[tabi].copy())
            out.tabheads.append(self.tabheads[tabi].copy())
        # copy filename
        out.filename=self.filename
        # return message and new object
        self.log.debug('Copy: done')
        return out

    def mergedata(self,other):
        """ Appends images and table from the new object to the existing
            data. The keywords from the primary headers are merged, in doubt
            the keyword from the local object is kept. The extension headers
            are left unchanged. If self has no data, the data from other is
            copied into the object.
            - other: a pipedata object with the same kind of data
            - raises ValueErrors if other has invalid format
        """
        ### check if other is correct data type
        if not isinstance(other,DataFits):
            self.log.error('Mergedata: other data is not PipeData type')
            raise TypeError('other data is not PipeData type')
        ### If self has no images and no tables: Copy All data from other
        if self.imgnames==[] and self.tabnames==[]:
            self=other.copy()
            return
        ### Check data compatibility
        # check if image number and names agree
        if self.imgnames != other.imgnames and self.imgnames != []:
            msg = 'other data has different image list'
            self.log.error('Mergedata: ' + msg)
            raise ValueError(msg)
        # check if table number and names agree
        if self.tabnames != other.tabnames:
            msg = 'other data has different table list'
            self.log.error('Mergedata: ' + msg)
            raise ValueError(msg)
        ### append the images (make new, copy, move)
        for imgi in range(len(self.imgdata)):
            # get new shape
            ssel=self.imgdata[imgi].shape
            soth=other.imgdata[imgi].shape
            # check dimensions, set new shape
            msg = 'dimension mismatch at image ' + self.imgnames[imgi]
            # if number of dimensions are equal
            if len(ssel) == len(soth):
                # check shape
                if ssel[1:] != soth[1:]:
                    self.log.error('MergeData: ' + msg)
                    raise ValueError(msg)
                nsel=ssel[0]
                noth=soth[0]
            # if number of other dimensions = number of self - 1
            elif len(ssel) == len(soth)+1:
                # check shape
                if ssel[1:] != soth:
                    self.log.error('MergeData: ' + msg)
                    raise ValueError(msg)
                nsel=ssel[0]
                noth=1
            # if dimensions are unmatched
            else:
                self.log.error('MergeData: ' + msg)
                raise ValueError(msg)
            nnew = nsel + noth
            snew = list(ssel)
            snew[0] = nnew
            # allocate new image
            newimg = numpy.empty(snew, dtype=self.imgdata[imgi].dtype)
            # copy images to new
            newimg[0:nsel, ...] = self.imgdata[imgi]
            newimg[nsel:nnew, ...] = other.imgdata[imgi]
            self.imgdata[imgi] = newimg
        ### append tables
        for tabi in range(len(self.tabdata)):
            # Check if table formats agree
            if self.tabdata[tabi].dtype != other.tabdata[tabi].dtype :
                msg='other table columns are incompatible'
                self.log.error('MergeData: '+msg)
                raise TypeError(msg)
            # make new shape
            nsel=len(self.tabdata[tabi])
            noth=len(other.tabdata[tabi])
            nnew=nsel+noth
            newshape=(nnew,)
            newtable=numpy.empty(newshape,dtype=self.tabdata[tabi].dtype)
            # copy existing table and put to self.data
            newtable[0:nsel]=self.tabdata[tabi]
            newtable[nsel:nnew]=other.tabdata[tabi]
            self.tabdata[tabi]=newtable
        ### combine headers
        self.mergehead(other)
        self.log.debug('MargeData: done')

    def mergehead(self, other):
        """ Merges the header of another data object to the existing header.
            Most of the header of the new data is ignored, COMMENT,
            HISTORY keywords are copied.
        """
        # get self and other cards and parents
        othercards = other.header.cards
        selfcards = self.header.cards
        # add history keywords (no duplicates)
        otherhist=[card.value for card in othercards
                  if card.keyword=='HISTORY'] # hist values for other
        selfhist=[card.value for card in selfcards
                  if card.keyword=='HISTORY'] # hist values for self
        for hist in otherhist: # loop through other history values
            if not hist in selfhist: # only add if not duplicate
                self.header.add_history(hist)
        # add comment keywords (no duplicates)
        othercomm=[card.value for card in othercards
                  if card.keyword=='COMMENT'] # comm values for other
        selfcomm=[card.value for card in selfcards
                  if card.keyword=='COMMENT'] # comm values for self
        for comm in othercomm: # loop through other comment values
            if not comm in selfcomm: # only add if not duplicate
                self.header.add_comment(comm)
        # Go through keywords listed in headmerge: assume self is first
        headmerge = self.config['headmerge']
        for key in headmerge.keys():
            if key in self.header and key in other.header:
                selfval = self.header[key]
                otherval = other.header[key]
                operation = headmerge[key].upper()
                if operation == 'LAST': selfval = otherval
                elif operation == 'MIN': selfval = min(selfval,otherval)
                elif operation == 'MAX': selfval = max(selfval,otherval)
                elif operation == 'SUM': selfval += otherval
                elif operation == 'OR': selfval = selfval | otherval
                elif operation == 'AND': selfval = selfval & otherval
                elif operation == 'CONCATENATE':
                    if ',' in selfval:
                        vlist = selfval.split(',')
                    else:
                        vlist = [selfval]
                    if otherval not in vlist:
                        vlist.append(otherval)
                        selfval = ','.join(sorted(vlist))
                elif operation == 'DEFAULT':
                    if type(selfval) is str:
                        selfval = 'UNKNOWN'
                    elif type(selfval) is int or type(selfval) is long:
                        selfval = -9999
                    elif type(selfval) is float:
                        selfval = -9999.0
                self.header[key] = selfval
        self.log.debug('MergeHead: done')

    def copyhead(self,other,name=None,overwrite=True):
        """ Similar to mergehead, except copy all header keywords, comments,
            and history from other.  Will overwrite existing cards, unless
            overwrite flag is set to False.  Exceptions: HISTORY and COMMENT
            cards are always appended to the end of the list of such keywords
            present in self.

            If name is None, use the header from the first HDU (self.header)
        """

        if name is None:
            head1 = other.header
            head2 = self.header
        else:
            head1 = other.getheader(name)
            head2 = self.getheader(name)
        try:
            nhist = len(head1['history'])
        except KeyError:
            nhist = 0
        try:
            ncomm = len(head1['comment'])
        except KeyError:
            ncomm = 0

        for k in head1.iterkeys():
            if k in ('COMMENT','HISTORY'): # handle later
                pass
            elif k in head2.keys():
                if overwrite:
                    head2[k] = (head1[k],head1.comments[k])
            else:
                head2[k] = (head1[k],head1.comments[k])
        for i in xrange(nhist):
            head2.add_history(head1['history'][i])
        for i in xrange(ncomm):
            head2.add_comment(head1['comment'][i])
        self.log.debug("Copydata: done")

    def adddata(self,images,tablerow):
        """ Appends images and a row to primary table
            - images: a dictionary with images to add
              (images are indexed by correct image names)
            - tablerow: one row of table values for the images (can be list)
                        This row is added to the first table
            - raises Errors if images or tablerow have wrong
              format or dimensions
        """
        # Comment: This function makes too many checks that would just
        #          raise exceptions when the error happens
        ### check input
        # check type of images
        if not isinstance(images, dict):
            self.log.error('Adddata: images needs to be a dictionary')
            raise TypeError('images needs to be a dictionary')
        # check keywords and data dimensions
        for key in images.keys():
            # check keyword / get index
            try:
                ind = self.imgnames.index(key.upper())
            except ValueError as error:
                msg = 'invalid image name (%s) in images' % key
                self.log.error('Adddata: ' + msg)
                raise error
            # check data type / get shape
            try:
                news = images[key].shape
            except AttributeError as error:
                msg = 'invalid image type for image %s' % key
                self.log.error('Adddata: ' + msg)
                raise error
            olds = self.imgdata[ind].shape
            # compare shapes (new can be part of old)
            if len(news) == len(olds):
                if news[1:] != olds[1:]:
                    msg = 'incompatible image shape in %s' % key
                    self.log.error('Adddata: ' + msg)
                    raise ValueError(msg)
            else:
                if news != olds[1:]:
                    msg = 'incompatible image shape in %s' % key
                    self.log.error('Adddata: ' + msg)
                    raise ValueError(msg)
        # check if table format is correct (by converting it)
        if len(tablerow) != len(self.table[0]) :
            self.log.error('AddData: table row has invalid length')
            raise ValueError('table row has invalid length')
        try:
            tablerow=numpy.rec.array(tablerow,dtype=self.table.dtype)
        except (ValueError,TypeError) as error:
            self.log.error('AddData: table row has invalid element type')
            raise ValueError('table row has invalid element type')
        ### append data
        for key in images.keys():
            # get index
            ind = self.imgnames.index(key.upper())
            # get new size
            news = images[key].shape
            olds = self.imgdata[ind].shape
            oldn = olds[0]
            if len(news) == len(olds):
                newn = news[0]
            else:
                newn = 1
            totn = oldn + newn
            tots = list(olds)
            tots[0]=totn
            # allocate new data
            newimg = numpy. empty(tots, dtype=self.imgdata[ind].dtype)
            # copy images and set new data
            newimg[0:oldn, ...] = self.imgdata[ind]
            newimg[oldn:totn, ...] = images[key]
            self.imgdata[ind] = newimg
        ### append table
        # make new shape
        totn = len(self.table) + 1
        newshape=(totn,)
        newtable=numpy.empty(newshape,dtype=self.table.dtype)
        # copy existing table and put to self.data
        newtable[0:totn-1]=self.table
        newtable[totn-1]=tablerow
        self.table=newtable
        self.log.debug('AddData: done')

    def copydata(self, other, dataname):
        """ Copies data (image or table) form another pipedata object to the
            current object. If an object of that name already exists it is
            overwritten. Both the data and the header are copied.
        """
        # Check type and presence of data in other object
        isimg = 0
        if dataname.upper() in other.imgnames:
            isimg = 1
            othind = other.imageindex(dataname)
        elif dataname.upper() in other.tabnames:
            othind = other.tableindex(dataname)
        else:
            self.log.error('CopyData: dataname not found')
            raise ValueError('dataname not found')
        # Check if name exists in local data, delete if it's different type
        if dataname.upper() in self.imgnames:
            if isimg < 1:
                self.imagedel(dataname)
        elif dataname.upper() in self.tabnames:
            if isimg > 0:
                self.tabledel(dataname)
        # Copy / Overwrite object
        if isimg>0:
            self.imageset(other.imgdata[othind].copy(), dataname,
                          other.imgheads[othind].copy())
        else:
            self.tableset(other.tabdata[othind].copy(), dataname,
                          other.tabheads[othind].copy())
        self.log.debug('CopyData: done')

    def imageindex(self,imagename=None):
        """ Returns the index of an image, such that the
            images can be accessed pipedata.imgdata[index]
            - imagename: The name of the requested image (None for first image)
        """
        # empty name -> return first image
        if imagename == None:
            self.log.debug('Imageindex: return 0 - done')
            return 0
        # check for valid name
        try:
            ind = self.imgnames.index(imagename.upper())
        except ValueError as error:
            msg = 'invalid image name (%s)' % imagename
            self.log.error('Imageindex: '+msg)
            raise error
        # return index
        self.log.debug('Imageindex: done')
        return ind

    def imageget(self,imagename=None):
        """ Returns an image
            - imgname: The name of the requested image (None for first image)
        """
        # get index
        ind = self.imageindex(imagename)
        # return image
        self.log.debug('Imageget: done')
        return self.imgdata[ind]

    def imageset(self, imagedata, imagename=None, imageheader=None, index=-1):
        """ Sets an image. This should be used to add a new image or if the
            image data should be replaced. The index flag allows to determine
            the position of the image in the image list.
            - imagedata: A multi dimensional array containing the image data
            - imagename: The name of the image to set (None for first image)
            - imageheader: The header for the image to set
            - index: Indicates the position of the image in the image list
                     Ignored if == -1.
        """
        totalindex = len(self.imgdata)
        # If index is not set, get a valid index
        if index < 0:
            if imagename == None:
                index = 0
                if totalindex > 0:
                    imagename = self.imgnames[0]
                else:
                    imagename = 'PRIMARY'
            elif imagename.upper() in self.imgnames:
                index = self.imgnames.index(imagename.upper())
            else: # name given, but not an existing name
                if totalindex > 0 and self.imgdata[0] is None:
                    index = 0 # add to primary HDU, if primary HDU empty
                else:
                    index = totalindex # add to end otherwise
        # re-use header, if none specified
        if imageheader is None:
            if index < totalindex:
                imageheader = self.imgheads[index]
        # Set image
        if index == 0:
            hdu = fits.PrimaryHDU(imagedata,header=imageheader)
        else:
            hdu = fits.ImageHDU(imagedata,header=imageheader)
        if index < totalindex: # overwriting an existing image
            self.imgnames[index] = imagename.upper()
            self.imgdata[index]  = imagedata
            self.imgheads[index] = hdu.header
        else:
            self.imgnames.append(imagename.upper())
            self.imgdata.append(imagedata)
            self.imgheads.append(hdu.header)
        self.log.debug('Imageset: done')

    def imagedel(self, imagename=None):
        """ Removes the image specified by imagename.
            - imagename: The name of the image to delete (None for first image)
        """
        # get index
        ind = self.imageindex(imagename)
        # delete image
        del self.imgnames[ind]
        del self.imgdata[ind]
        del self.imgheads[ind]
        self.log.debug('Imagedel: done')

    def tableindex(self, tablename=None):
        """ Returns the index of a table, such that the tables can be
            accessed pipedata.tabdata[index]
            - tablename: The name of the requested table (None for first table)
        """
        # Check if tables are present
        if len(self.tabnames) == 0:
            msg = 'no tables in data'
            self.log.error('Imageindex: ' + msg)
            raise RuntimeError(msg)
        # empty name -> return first table
        if tablename == None:
            self.log.debug('Tableindex: return 0 - done')
            return 0
        # check for valid name
        try:
            ind = self.tabnames.index(tablename.upper())
        except ValueError as error:
            msg = 'invalid table name (%s)' % tablename
            self.log.error('Tableindex: '+msg)
            raise error
        # return index
        self.log.debug('Tableindex: done')
        return ind

    def tableget(self, tablename = None):
        """ Returns a table
            - tablename: The name of the requested table (None for first table)
        """
        # get index
        ind = self.tableindex(tablename)
        # return table
        self.log.debug('Tableget: done')
        return self.tabdata[ind]

    def tableset(self, tabledata, tablename = None, tableheader = None,
                 index = -1):
        """ Sets a table. This should be used to add a new image or if the
            table data should be replaced. The index flag allows to determine
            where (in the order of tables) the table should be.
            - tabledata: A record array with table contents
            - tablename: The name of the table to set (None for first table)
            - index: Indicates the position of the table in the table list
                     Ignored if == -1.
        """
        # If index is not set, get valid index
        if index < 0:
            if tablename == None:
                index = 0
            # if image exists - replace
            elif tablename.upper() in self.tabnames:
                index = self.tabnames.index(tablename.upper())
        # Set table
        if index > len(self.tabnames)-1 or index < 0:
            self.tabnames.append(tablename.upper())
            self.tabdata.append(tabledata)
            # make sure tableheader is valid
            if tableheader == None:
                tableheader = fits.Header()
            self.tabheads.append(tableheader)
        else:
            self.tabnames[index] = tablename.upper()
            self.tabdata[index] = tabledata
            if tableheader != None:
                self.tabheads[index] = tableheader
        self.log.debug('Tableset: done')

    def tabledel(self, tablename = None):
        """ Removes the table specified by tablename.
            - tablename: The name of the table to be deleted        # Remove Imag columns
        #for tabcol in ['R array Imag','T array Imag','Chop Offset Imag']:
        #    self.dataout.table.columns.del_col(tabcol)

        # Remove Imag columns
        #for tabcol in ['R array Imag','T array Imag','Chop Offset Imag']:
        #    self.dataout.table.columns.del_col(tabcol)

        # Remove Imag columns
        #for tabcol in ['R array Imag','T array Imag','Chop Offset Imag']:
        #    self.dataout.table.columns.del_col(tabcol)

        # Remove Imag columns
        #for tabcol in ['R array Imag','T array Imag','Chop Offset Imag']:
        #    self.dataout.table.columns.del_col(tabcol)

                         (None for first table)
        """
        # get index
        ind = self.tableindex(tablename)
        # delete table
        del self.tabnames[ind]
        del self.tabdata[ind]
        del self.tabheads[ind]
        self.log.debug('Tabledel: done')

    def tableaddcol(self, colname, array, tablename = None, dtype=None):
        """ Adds a column to the table. If the table under tablename
            doesn't exist, it is created.
            - colname: a string specifying the name of the new column
            - array: a list containing the values for the new column
            - tablename: The name of the affected table (None for first table)
            - dtype: (optional) data type for the new column
            WARNING: This may not work if the elements in the table have more
                     than one dimension.
        """
        ### Make new data type
        # make a basic array
        array=numpy.asarray(array)
        # set correct data type
        if dtype is None:
            dtype=array.dtype
        # get additional dimension
        if len(array.shape) > 1:
            newtype = ( colname, dtype, array.shape[1] )
        else:
            newtype = ( colname, dtype )
        ### Get table index (-1 if table needs to be created)
        tabind = -1
        if tablename is None:
            if len(self.tabnames) > 0:
                tabind = 0
            else:
                tablename = "Table"
        else:
            if tablename.upper() in self.tabnames:
                tabind = self.tableindex(tablename)
        ### Make new table if necessary
        if tabind < 0:
        # If there is no existing table:
            # Get new table data type
            newdtype = numpy.dtype([newtype])
            # Make new table
            newtable = numpy.empty(array.shape[0],dtype=newdtype)
            # Add to list of tables
            tabind = len(self.tabnames)
            self.tabdata.append(newtable)
            self.tabnames.append(tablename.upper())
            self.tabheads.append(fits.Header())
        else:
        ### Else (table exists) base new table on old table
            table = self.tabdata[tabind]
            # check if dimension of new value array is correct
            if len(table) != len(array):
                msg = 'column array len (%d) != table len (%d)' % (len(array), len(table))
                self.log.error('TableAddCol: ' + msg)
                raise ValueError(msg)
            # get new data type for table
            newdtype = numpy.dtype(table.dtype.descr+[newtype])
            # make new table
            newtable = numpy.empty(table.shape,dtype=newdtype)
            # fill old table values
            for field in table.dtype.fields:
                newtable[field] = table[field]
        # fill new table values
        newtable[colname] = array
        # copy new table to self.table
        self.tabdata[tabind] = newtable
        self.log.debug('TableAddCol: done')

    def tableaddrow(self, tablerow, tablename = None):
        """ Adds a row to the data table
            - tablerow: A list containing the elements of the row to be added
            - tablename: The name of the affected table (None for first table)
        """
        # Get table
        tabind = self.tableindex(tablename)
        table = self.tabdata[tabind]
        # check if tablerow format is correct (by converting it)
        if len(tablerow) != len(table[0]) :
            self.log.error('TableAddRow: table row has invalid length')
            raise ValueError('table row has invalid length')
        try:
            tablerow=numpy.rec.array(tablerow,dtype=table.dtype)
        except (ValueError,TypeError):
            self.log.error('TableAddRow: table row has invalid element type')
            raise ValueError('table row has invalid element type')
        # Add to table
        self.tabdata[tabind] = numpy.insert(table, len(table), tablerow)
        self.log.debug('TableAddRow: done')

    def tabledelcol(self, colname, tablename = None):
        """ Deletes a column of the data table.
            - colname: A string or list of strings containing the
                       name(s) of the column(s) to delete
            - tablename: The name of the affected table
                         (None for first table)
        """
        # Get table
        tabind = self.tableindex(tablename)
        table = self.tabdata[tabind]
        # Check if colname is valid, make sure it's a list
        olddt = table.dtype
        if isinstance(colname,str):
            colname = [colname]
        if isinstance(colname,(list,tuple)):
            for c in colname:
                if c not in olddt.names:
                    msg = 'Invalid column name %s' % c
                    self.log.error('TableDelCol: ' + msg)
                    raise ValueError(msg)
        else:
            msg = "Invalid colname '%s'.  Must be string or list/tuple" %colname
            self.log.error('TableDelCol: ' + msg)
            raise ValueError(msg)
        # if all columns deleted, clear table
        if len(olddt) - len(colname) <= 0:
            del self.tabdata[tabind]
            del self.tabnames[tabind]
            del self.tabheads[tabind]
        # else -> remove column(s) from table
        else:
            # if it's a pyfits.FITS_rec -> Treat specially
            # (this was pyfits.fitsrec.FITS_rec)
            # test for FITS_rec.  ndarray and recarray don't support names method
            try:
                names   = table.names
                names   = table.names # list of fields
                formats = table.columns.formats
                dims    = table.columns.dims
                units   = table.columns.units

                cols = []
                for n,f,d,u in zip(names,formats,dims,units):
                    if n not in colname:
                        cols.append(fits.Column(name=n,format=f,dim=d,unit=u,
                            array=table.field(n)))
                tbhdu = fits.BinTableHDU.from_columns(fits.ColDefs(cols))
                self.tabdata[tabind] = tbhdu.data
                self.tabheads[tabind] = tbhdu.header
            except AttributeError: # assume table is a regular record array
                newnames = [n for n in olddt.names if n not in colname]
                newtable = table[newnames]
                self.tabdata[tabind] = newtable

        self.log.debug('TableDelCol: done')

    def tabledelrow(self, index, tablename = None):
        """ Deletes a row of the data table.
            - index: The index of the row to delete
            - tablename: The name of the affected table (None for first table)
        """
        # Get table
        tabind = self.tableindex(tablename)
        table = self.tabdata[tabind]
        # Check if index is valid
        if index >= len(table):
            self.log.error('TableDelRow: Invalid row index %d' % index)
            raise ValueError('Invalid row index %d' % index)
        # if table has one row -> clear table
        if len(table) < 2:
            del self.tabdata[tabind]
            del self.tabnames[tabind]
        # else -> remove row from table
        else:
            # make index list
            #indlist = [i for i in range(len(table)) if i != index]
            # fill in the values
            #newtable = table[indlist]
            # copy new table
            #self.tabdata[tabind] = newtable.copy()
            self.tabdata[tabind] = numpy.delete(table,index).copy()
        self.log.debug('TableDelRow: done')

    def tablemergerows(self, rows):
        """ Merges several table rows into a single row that is returned.
            Each column is merged according to the rules defined in the
            [table] section of the configuration file.

            Usage: - Get lines to be merged from input table
                     Ex: inrows = indata.table[10:20]
                   - Call tablemergerows
                     Ex: outrow = indata.tablemergerows(inrows)
                   - Add / Insert row to output table
                     Ex: outdata.tableaddrow(outrow)
        """
        # check if rows has same format than table
        try:
            _newdtype = rows.dtype
        except AttributeError as error:
            msg = 'input rows are incorrect data type'
            self.log.error('TableMergeRows: %s' % msg)
            raise error
        #if newdtype != table.dtype :
        #    msg = 'input rows have different format than table'
        #    self.log.error('TableMergeRows: %s' % msg)
        #    raise TypeError(msg)
        # make output table row (copy from first row)
        outrow = (rows[0:1].copy())[0]
        # run through columns and merge values
        for colname in rows.dtype.names:
            # get merge function
            try:
                funct = self.config['table'][colname.lower()].lower()
            except :
                # if keyword is not available
                self.log.warn('TableMergeRows: Missing table merge entry for '
                    + 'column -%s- returning first row value' % colname)
                funct = 'first'
            # Try to run the function
            # Comment: float() is necessary, otherwise values get messed up
            try:
                if funct == 'first':
                    pass
                elif funct == 'last':
                    outrow[colname] = rows[-1][colname]
                elif funct == 'min':
                    outrow[colname] = float(numpy.nanmin(rows[colname]))
                elif funct == 'max':
                    outrow[colname] = float(numpy.nanmax(rows[colname]))
                elif funct == 'med':
                    outrow[colname] = float(numpy.nanmedian(rows[colname]))
                elif funct == 'avg':
                    outrow[colname] = float(numpy.nanmean(rows[colname] ))
                elif funct == 'sum':
                    outrow[colname] = float(numpy.nansum(rows[colname]))
                elif funct == 'wtavg':
                    tmp = float(numpy.nansum(rows[colname]*rows['Samples']))
                    tmpsum = float(numpy.nansum(rows['Samples']))
                    if tmpsum > 0:
                        outrow[colname] = tmp/tmpsum
                    else:
                        outrow[colname] = tmp*0.0
                else:
                    self.log.warn('TableMergeRows: Unknown operation -'+funct
                     +'- for column -'+colname+'- returning first row value')
            except NameError:
                # if unsuccessful return error
                self.log.warn('TableMergeRows: Error in %s( %s ) - '
                    % (funct, colname) + ' returning first row value')
        # return output row
        self.logcount += 1
        if self.logcount%10 == 0:
            self.log.debug('TableMergeRows: done')
        return outrow

    def tablemergetables(self,tables):
        """Returns a new table containing data merged from the input table(s).
           Columns are merged according to the rules define in the [table]
           section of the configuration file.

           tables is a list of tables, previously obtained with something
           like tableget().  Note that each table is assumed to have a single
           row of data.  If you need to merge rows of data, use something
           like tablemergerows().

        """

        ntable = len(tables)
        self.log.debug("ntable = %d" %ntable)
        names   = tables[0].names # list of fields
        formats = tables[0].columns.formats
        dims    = tables[0].columns.dims
        units   = tables[0].columns.units
        # loop through all tables, make sure names, formats, dims, and units
        # are the same among them.
        for i in xrange(1,ntable):
            if cmp(names,tables[i].names): # equals zero if identical
                self.log.error('column names differ for merging')
                raise ValueError
            if cmp(formats,tables[i].columns.formats): # equals zero if identical
                self.log.error('column formats differ for merging')
                raise ValueError
            if cmp(dims,tables[i].columns.dims): # equals zero if identical
                self.log.error('column names differ for merging')
                raise ValueError
            if cmp(units,tables[i].columns.units): # equals zero if identical
                self.log.error('column names differ for merging')
                raise ValueError

        cols = []
        for n,f,d,u in zip(names,formats,dims,units):
            # get merge function
            try:
                funct = self.config['table'][n.lower()].lower()
            except :
                # if keyword is not available
                self.log.warn('TableMergeTables: Missing table merge entry for '
                    + 'column -%s- returning first row value' % n)
                funct = 'first'
            if funct == 'first':
                tmp = tables[0][n]
            elif funct == 'last':
                tmp = tables[-1][n]
            elif funct == 'min':
                tmp = numpy.nanmin([a[n] for a in tables])
            elif funct == 'max':
                tmp = numpy.nanmax([a[n] for a in tables])
            elif funct == 'med': # median
                tmp = numpy.nanmedian([a[n] for a in tables])
            elif funct == 'avg': # average
                tmp = numpy.nanmean([a[n] for a in tables])
            elif funct == 'sum':
                tmp = numpy.nansum([a[n] for a in tables])
            elif funct == 'wtavg':
                tmp = numpy.nansum([a[n]*a['Samples'] for a in tables])
                tmpsum = numpy.nansum([a['Samples'] for a in tables])
                tmp = tmp/float(tmpsum)
            else:
                self.log.warn('TableMergeTables: Unknown operation -'+funct
                    +'- for column -'+n+'- returning first row value')
                tmp = tables[0][n]
            cols.append(fits.Column(name=n, format=f, dim=d, unit=u,
                array=[tmp]))

        tbhdu = fits.BinTableHDU.from_columns(fits.ColDefs(cols))
        return tbhdu

    def getheader(self, dataname = ''):
        """ Returns the stored header. If dataname is specified, the
            header of the named image/table is returned, otherwise the
            primary header.
        """
        # Return primay header
        if dataname == '':
            header = self.header
        # Return image header
        elif dataname.upper() in self.imgnames:
            index = self.imageindex(dataname)
            header = self.imgheads[index]
        # Return table header
        elif dataname.upper() in self.tabnames:
            index = self.tableindex(dataname)
            header = self.tabheads[index]
        # Dataname not found -> return error message
        else:
            msg = 'Invalid data name (%s)' % dataname
            self.log.error('GetHeader: ' + msg)
            raise ValueError(msg)
        self.log.debug('GetHeader: done')
        return header

    def setheader(self,header, dataname = ''):
        """ Overwrites the stored header. If dataname is specified, the
            header of the named image/table is set, otherwise the
            primary header.
        """
        # Set primay header
        if dataname == '':
            self.header = header
        # Set image header
        elif dataname.upper() in self.imgnames:
            index = self.imageindex(dataname)
            self.imgheads[index] = header
        # Set table header
        elif dataname.upper() in self.tabnames:
            index = self.tableindex(dataname)
            self.tabheads[index] = header
        # Dataname not found -> return error message
        else:
            msg = 'Invalid data name (%s)' % dataname
            self.log.error('SetHeader: ' + msg)
            raise ValueError(msg)
        self.log.debug('SetHeader: done')


    def getheadval(self, key, dataname = '', errmsg=True):
        """ Get Header Value: Returns the value of the requested key from
            the header. If the keyword is present in the [Header] section
            of the configuration that value is returned instead. In case that
            value from the configuration file is itself a header key, the value
            stored under that key is returned. If the key can not be found an
            KeyError is produced and a warning is issued.
        """
        val = None
        inkey = key # retain key which was input in case key changes
        # Look in the config
        try:
            # get the value
            val = self.config['header'][key.upper()]
            # Check if it's optional header replacement i.e. starts with '?_'
            if val[:2] in ['?_', '? ', '?-']:
                # if key is not in the header -> use key name under value instead
                if key not in self.header:
                    self.log.info('Getheadval: Using %s keyword for %s' %
                                  (val[2:].upper(), key))
                    key = val[2:].upper()
                val = None
            # Check if it's a Header replacement (but not T/F)
            elif val[0].isupper() and val[:2] not in ['T ', 'F ']:
                self.log.info('Getheadval: Using %s value for %s' %
                              (val.upper(), key ))
                key = val.upper()
                val = None
            else:
                # make it a pyfits.Card then get value and comment
                card = fits.Card()
                card = card.fromstring(key.upper() + ' = ' + val)
                # Following code is unused because non-string values
                #   don't get converted correctly. However the new code
                #   may not work with all version of pyfits - tbd - nlc 2013
                #tmp = val.split('/')
                #if len(tmp) > 1:
                #    card=pyfits.Card(key,tmp[0].strip(),tmp[1].strip())
                #else:
                #    card=pyfits.Card(key,tmp[0].strip())
                self.log.info('Getheadval: Setting %s to %s' % (key, val))
                val=card.value
                # update value in header
                self.setheadval(key, val, card.comment)
        except KeyError :
            # if key is not in config - continue
            pass
        except TypeError :
            # if config is not yet loaded - return error
            self.log.warn('GetHeadVal: Missing Configuration')
            # The following line is commented out such that pipedata can
            #     still be used without configuration
            #raise RuntimeError('Missing Pipe Configuration')
        # Look in the header
        if val is None:
            # get the header from dataname
            header = self.getheader(dataname)
            # get value from header
            try:
                val = header[key]
            except KeyError:
                # if keyword is not found
                msg='Missing %s keyword in header %s' % (key, dataname)
                if errmsg:
                    self.log.error('GetHeadVal: %s' % msg)
                raise KeyError(msg)
        # if Value is a pyfits.core.Undefined i.e. no keyword
        if isinstance(val, fits.Undefined):
            msg = 'Missing value for key = %s - returning empty string' % key
            self.log.warn('GetHeadVal: %s' % msg)
            val = ''
        # Debug message (only first line if multi-line value)
        if '\n' in str(val):
            self.log.debug('GetHeadval: done (%s=%s ...)' % (inkey, str(val[0])))
        else:
            self.log.debug('GetHeadVal: done (%s=%s)' % (inkey,str(val)))
        return val

    def setheadval(self, key, value, comment=None, dataname = ''):
        """ Sets a FITS keyword in the current header
        """
        # If no header exitst, make a first empty image
        if len(self.imgheads) == 0:
            self.imgheads = [fits.PrimaryHDU().header]
            self.imgdata = [None]
            self.imgnames = ['PRIMARY']
        # Set primay header
        if dataname == '':
            hdr = self.header
        # Set image header
        elif dataname.upper() in self.imgnames:
            index = self.imageindex(dataname)
            hdr = self.imgheads[index]
        # Set table header
        elif dataname.upper() in self.tabnames:
            index = self.tableindex(dataname)
            hdr = self.tabheads[index]
        # Dataname not found -> return error message
        else:
            msg = 'Invalid data name (%s)---' % dataname
            self.log.error('SetHeadVal: ' + msg)
            raise ValueError(msg)
        # Set value into hdr
        if comment != None: hdr[key] = (value,comment)
        else: hdr[key] = value
        self.log.debug('SetHeadVal: done')

    def delheadval(self, key, dataname = None):
        """ Delete one or more FITS keyword in specified header, which defaults
            to the first header.
        """

        if isinstance(key,(list,tuple)):
            for k in key:
                self.delheadval(k,dataname)
        elif isinstance(key,str):
            if dataname is None:
                try:
                    del(self.header[key])
                except KeyError: # don't care if it doesn't exist
                    pass
            elif dataname.upper() in self.imgnames:
                index = self.imageindex(dataname)
                try:
                    del(self.imgheads[index][key])
                except KeyError: # don't care if it doesn't exist
                    pass
            elif dataname.upper() in self.tabnames:
                index = self.tableindex(dataname)
                try:
                    del(self.tabheads[index][key])
                except KeyError: # don't care if it doesn't exist
                    pass
            else: # Dataname not found -> return error message
                msg = 'Invalid data name (%s)---' % dataname
                self.log.error('DelHeadVal: ' + msg)
                raise ValueError(msg)
        else:
            msg = 'Invalid key (%s).  Must be a str, list, or tuple' %repr(key)
            self.log.error('DelHeadVal: ' + msg)
            raise ValueError(msg)
        self.log.debug('DelHeadVal: done')

    def test(self):
        """ Test Pipe Data Object: Runs a set of tests on a pipedata object.
            This function needs to start in the folder contining the test
            files.
        """
        self.log.info("Testing Data Object")
        ### Setup
        # test loading config
        #self.setconfig(1) # should raise TypeError
        self.setconfig(None)
        conf = self.testconf
        self.setconfig(conf)
        conf=configobj.ConfigObj(conf)
        self.setconfig(conf)
        ### Loading and Saving DAta
        # self.load(1) # should create TypeError
        # self.load('nothere.fits') # should create IOError
        # self.load('makebasic.pro') # should create IndexError
        testpath = self.config['testing']['testpath']
        print(testpath+'/makedata/testimgtabimg.fits')
        self.load(testpath+'/makedata/testimgtabimg.fits')
        self.log.debug('Images = ' + repr(self.imgnames))
        self.save(testpath+'/testsave.fits')
        print(self.filenamebegin, self.filenameend)
        #self.save(1) # creates TypeError
        ### Copy, Merge and Add Data
        other=self.copy()
        other.mergedata(self)
        other.load(testpath+'/makedata/testimgtabimg.fits')
        other.imgdata[1]=numpy.zeros((10,20))
        #other.table = numpy.rec.array([(1)], names='nr', formats='i2')
        self.mergedata(other)
        self.adddata({'First Image':numpy.zeros((10,10))},['Delta',4])
        for i in range(len(self.imgdata)):
            s='%s %s' % (self.imgnames[i],self.imgdata[i].shape)
            self.log.debug(s)
        self.log.debug(repr(self.table))
        ### Images / Table / Header
        # Images
        img=self.imageget('SECOND IMAGE')
        self.imageset(numpy.zeros((5,5,5)),'First Image Copy')
        self.imagedel('SECOND IMAGE')
        self.imageset(img,'Second Image',index=-1)
        other.imgnames[0]='WILDCARD'
        self.copydata(other,'WILDCARD')
        # Image variable
        self.image=numpy.zeros((5,5))
        self.image[0,:]=range(5)
        # Table
        i=len(self.table)
        #print(self.table,self.table.dtype,self.table.shape)
        self.tableaddcol('index',numpy.arange(i)-2)
        self.tableaddrow(['Zeta',0,100])
        self.tabledelcol('NR')
        self.tabledelrow(3)
        #print(self.table,self.table.dtype,self.table.shape)
        self.table['index']=numpy.arange(len(self.table))+2
        self.table[0]=('first',100)
        self.tableaddcol('arr', numpy.ones((7,2)))
        self.log.debug(repr(self.table))
        self.log.debug(repr(self.table.dtype))
        self.log.debug('TabMrg='+repr(self.tablemergerows(self.table)))
        # Second table
        tab2 = self.tableget()
        self.tableset(tab2,'Second Table')
        print(self.tableget('Second Table').dtype,self.tabdata[1])
        self.tabledelcol('arr','Second Table')
        self.tabledelrow(6,'Second Table')
        self.tabledelrow(5,'Second Table')
        self.tableaddrow(['velo',19], 'Second Table')
        self.tableaddcol('NewNr', range(6), 'Second Table')
        print(self.tableget('Second Table').dtype,self.tabdata[1])
        # Delete first table and save
        self.tabledel('TABLE')
        self.save(testpath+'/testsave.fits')
        # Header
        self.header['Reduced'] = 0
        self.setheadval('TST','Test Value','') #,'Second Table')
        print(self.header)
        self.getheadval('TST') #,'Second Table')
        self.setheadval('TST',15)
        print(self.header)
        self.setheader(self.getheader())
        # Additional Load and Save Tests
        if os.path.exists(testpath+'/makedata/outtest.fits'):
            os.remove(testpath+'/makedata/outtest.fits')
        self.save(testpath+'/makedata/outtest.fits')
        #self.save('pipedata/outtest.fits')
        #self.load('pipedata/outtest.fits')
        #self.load('sample.chop_nod.dog.raw.fits')
        #self.load('sample.noise.fft.raw.fits')
        self.log.info("Test Finished")
        print('Done')
        return 0

if __name__ == '__main__':
    """ Main Function for testing:
        Changes to the folder with the test files and runs pipedata.test()
    """
    logging.basicConfig(level=logging.DEBUG)
    data=DataFits()
    data.test()

""" === History ===
    2016-9-26 Marc Berthoud: Split PipeData (fits only) into
        DataParent and DataFits
        Main Changes:
        - Added data attribute which can go to image or table
    2015-9-25 Marc Berthoud: Edited save() to update keywords for
        FILENAME and DATE
    2014-7-31 Marc Berthoud: Added output_verify=fix to write()
        Fixed tablemergerows to work for FITS_rec
        Now use numpy.insert for tableaddrow, numpy.delete in tabledelrow
    2014-2-17 Marc Berthoud: Fix error with load if primary header empty
    2012-11-15 Marc Berthoud: Loadhead now loads from other HDUs
    2012-9-27 Marc Berthoud: Add functionality for header keyword replacement
    2012-9-13 Marc Berthoud: Allow files to have empty HDUs
        i.e. image.shape=(0,)
    2012-5-17 Marc Berthoud: Added storate and management of headers for all
    2012-2-17 Marc Berthoud: Added multiple tables option
        Many new commands (tableindex/get/set/del) and upgrades to table
        commands.
        COMMENT: adddata now only affects first table (may have to change)
    2011-2-4 Marc Berthoud: Added filenamebase to __getattr__
    2010-12-20 Marc Berthoud: Made sure that all image names would always be
        uppercase.
    2010-12-3 Marc Berthoud: Added code to tableaddcol such that arrays
        can be stored in tables.
    2010-10-1  Marc Berthoud, Ver0.1.1: New structure for images:
        - Now there can be multiple images
        - The size of the primary image doesn't have to agree with the
          table size
    2009-11-17 Marc Berthoud, Added remap to sky index
    2008-10-28 Marc Berthoud, Ver0.1: Wrote and tested
"""



""" PIPE STEP No-Input PARENT

    This module defines a pipeline step with no input file (i.e. self.datain is
    ignored).  It creates a single self.dataout pipedata object.  This module
    is a child of StepMOParent.
    
    (We are the Knights who say NI!)
    
    @author: chapman
"""

import logging # logging object library
from configobj import ConfigObj
from .dataparent import DataParent #pipeline data object
from .stepmoparent import StepMOParent # pipe step parent object

class StepNIParent(StepMOParent):
    """ Pipeline no data input and multiple data output parent object
    """
    
    stepver = '1.0' # pipe step version
    
    def __init__(self):
        """ Constructor: Initialize data objects and variables
            calls the setup function.
        """
        # call superclass constructor (calls setup)
        super(StepNIParent,self).__init__()
        # Change dataout
        self.dataout = [DataParent()]
        # set iomode
        self.iomode = 'NIMO'

    def setup(self):
        """ ### Names and Parameters need to be Set Here ###
            Sets the internal names for the function and for saved files.
            Defines the input parameters for the current pipe step.
            Setup() is called at the end of __init__
            The parameters are stored in a list containing the following
            information:
            - name: The name for the parameter. This name is used when
                    calling the pipe step from command line or python shell.
                    It is also used to identify the parameter in the pipeline
                    configuration file.
            - default: A default value for the parameter. If nothing, set
                       '' for strings, 0 for integers and 0.0 for floats
            - help: A short description of the parameter.
        """
        ### Set Names
        # Name of the pipeline reduction step
        self.name='parentni'
        # Shortcut for pipeline reduction step and identifier for
        # saved file names.
        self.procname = 'unk'
        # Set Logger for this pipe step
        self.log = logging.getLogger('pipe.step.%s' % self.name)
        ### Set Parameter list
        # Clear Parameter list
        self.paramlist = []
        # Append parameters
        self.paramlist.append(['filloutput', False,
            'If TRUE the step will return one pipedata object - else an empty list is returned (default)'])
        
    def __call__(self,datain=None, **arglist):
        """ Object Call: returns reduced input data
            No datain is allowed, in this case the step expects to receive
            the configuration in another way Ex: step.config=config
        """
        # Get input data
        self.datain = datain
        # Start Setup
        self.runstart(self.datain, arglist)
        # Call the run function
        self.run()
        # Finish - call end
        self.runend(self.dataout)
        # return result
        return self.dataout

    def runstart(self, data, arglist):
        """ Method to call at the beginning of the pipe step call.
            - Sends initial log message
            - Checks the validity of the input data
            - Gets configuration from input data and checks type
        """
        # Start Message
        self.log.info('Start Reduction: Pipe Step %s' % self.name)
        # Set input arguments
        for k in arglist.keys():
            self.arglist[k.lower()] = arglist[k]
        # Check input data type and set data config
        if issubclass(data.__class__, DataParent):
            self.config = data.config
        elif self.config == None:
            msg='Invalid configuration information - aborting'
            self.log.error(msg)
            raise RuntimeError('Runstart: '+msg)
        # Check configuration
        if not isinstance(self.config,ConfigObj):
            msg='Invalid configuration information - aborting'
            self.log.error(msg)
            raise RuntimeError('Runstart: '+msg)
    
    def run(self):
        """ Runs the data reduction algorithm. The result is in self.dataout.
            NI steps a NIMO so multiple outputs; they have to return a list
            of data objects.
        """
        # Return dataout
        if self.getarg('filloutput'):
            # Make a datafits object and put it into output
            dout = DataParent(config = self.config)
            dout.filename = 'StepNIOuput.fits'
            self.dataout = [dout]
        else:
            # Return empty list
            self.dataout = []

    def execfiles(self, inputfiles):
        """ Runs the step without an input file
        """
        #self.datain = DataParent(config = self.config)
        # Call start - run and call end
        #self.runstart(self.datain,self.arglist)
        self.runstart(None, self.arglist)
        self.run()
        self.runend(self.dataout)
        # Write output file
        if len(self.dataout)>0:
            for data in self.dataout:
                data.save()
                self.log.info('Execute: Saved result %s' % data.filename)
        
    def test(self):
        """ Test Pipe Step Parent Object:
            Runs a set of basic tests on the object
        """
        # log message
        self.log.info('Testing pipe step parent')
        # test function call
        #testout=self(1) # should raise TypeError
        if self.config != None:
            testin = DataParent(config=self.config)
        else:
            testin = DataParent(config=self.testconf)
        testin.filename = 'this.file.type.fts'
        testout=self(testin, sampar=5.0)
        print(testout.header)
        print(testout.filename)
        # test get and set parameters
        print("sampar=%.3f" % self.getarg('sampar'))
        # log message
        self.log.info('Testing pipe step parent - Done')

if __name__ == '__main__':
    """ Main function to run the pipe step from command line on a file.
        Command:
          python stepniparent.py input.fits -arg1 -arg2 . . .
        Standard arguments:
          --config=ConfigFilePathName.txt : name of the configuration file
          -t, --test : runs the functionality test i.e. pipestep.test()
          --loglevel=LEVEL : configures the logging output for a particular level
          -h, --help : Returns a list of 
    """
    StepNIParent().execute()

""" === History ===
    2018-07-12 Matt Merz: Updated and tested version 1.0
    2014-12-30 Nicholas Chapman: Written and tested version 0.1
"""



# All this will allow to from darepype.drp import DataParent
#   instead of from darepype.drp.dataparent import DataParent
from .dataparent import DataParent
from .datafits import DataFits
from .stepparent import StepParent
from .stepmiparent import StepMIParent
from .stepmoparent import StepMOParent
from .stepniparent import StepNIParent
from .pipeline import PipeLine



''' PIPE STEP Multi-Input PARENT

    This module defines a pipeline step with multiple input data objects
    and one output data object. All such pipe steps are descendants of
    this one.

    @author: berthoud
'''

import os
import re
import logging # logging object library
from .dataparent import DataParent # pipeline data object
from .stepparent import StepParent # pipe step parent object

class StepMIParent(StepParent):
    """ Pipeline multiple data input parent object
    """

    stepver = '1.0' # pipe step version

    def __init__(self):
        """ Constructor: Initialize data objects and variables
            calls the setup function.
        """
        # call superclass constructor (calls setup)
        super(StepMIParent,self).__init__()
        # Change datain
        self.datain = [DataParent()]
        # set iomode
        self.iomode = 'MISO'
        # add a filenum list, for output filenames
        self.filenum = []

    def setup(self):
        """ ### Names and Prameters need to be Set Here ###
            Sets the internal names for the function and for saved files.
            Defines the input parameters for the current pipe step.
            Setup() is called at the end of __init__
            The parameters are stored in a list containing the following
            information:
            - name: The name for the parameter. This name is used when
                    calling the pipe step from command line or python shell.
                    It is also used to identify the parameter in the pipeline
                    configuration file.
            - default: A default value for the parameter. If nothing, set
                       '' for strings, 0 for integers and 0.0 for floats
            - help: A short description of the parameter.
        """
        ### Set Names
        # Name of the pipeline reduction step
        self.name='parentmi'
        # Shortcut for pipeline reduction step and identifier for
        # saved file names.
        self.procname = 'unk'
        # Set Logger for this pipe step
        self.log = logging.getLogger('pipe.step.%s' % self.name)
        ### Set Parameter list
        # Clear Parameter list
        self.paramlist = []
        # Append parameters
        self.paramlist.append(['sampar', 1.0,
            'Sample Parameter - parent only - no practical use'])

    def run(self):
        """ Runs the data reduction algorithm. The self.datain is run
            through the code, the result is in self.dataout.
        """
        # Log the value of sample parameter
        self.log.debug("Sample Parameter = %.2f" % self.getarg('sampar'))
        # Return the first datain element
        self.dataout = self.datain[0]

    def runstart(self, data, arglist):
        """ Method to call at the beginning of the pipe step call.
            Mostly refers to runstart of stepparent
        """
        # Keep a list of input file numbers for output filename
        self.filenum = []

        # Check input data - should be a list/tuple with PipeData objects
        if isinstance(data, (list,tuple)):
            for d in data:
                if not isinstance(d, DataParent):
                    msg = 'Invalid input data type: Pipe Data object is required'
                    self.log.error(msg)
                    raise TypeError('Runstart: '+msg)
                # try to read numerical file number from input name
                try:
                    # test if it is a valid number
                    fnum = int(d.filenum)
                    # append the string version if it is
                    self.filenum.append(d.filenum)
                except (ValueError, TypeError):
                    pass
        else:
            msg = 'Invalid input data type: List object is required'
            self.log.error(msg)
            raise TypeError('Runstart: '+msg)
        # Call parent runstart
        super(StepMIParent,self).runstart(data[0],arglist)

    def updateheader(self,data):
        """ Update the header for a single PipeData object
            - Sets the PROCSTAT and PROCLEVL keywords in the data header
            - Adds a history entry to the data header
            - Update the output filename
            Mostly refers to updateheader of stepparent
        """
        # Call parent updateheader
        super(StepMIParent,self).updateheader(data)

        # Update file name with PipeStepName and input filenumbers
        # if available and MISO. Otherwise, use the version set by the parent.
        if self.iomode == 'MISO' and len(self.filenum) > 1:
            fn = sorted(self.filenum)
            filenums = fn[0] + '-' + fn[-1]
            outdir, basename = os.path.split(data.filename)
            match = re.search(self.config['data']['filenum'],basename)
            if match is not None:
                # regex may contain multiple possible matches --
                # for middle or end of filename
                for i,g in enumerate(match.groups()):
                    if g is not None:
                        fbegin = basename[:match.start(i+1)]
                        fend = basename[match.end(i+1):]
                        data.filename = os.path.join(outdir,
                                                     fbegin + filenums + fend)
                        break            
            #fileend = '_' + fn[0] + '-' + fn[-1] + '.fits'
            #data.filename = data.filenamebegin + self.procname.upper() + fileend

    def execfiles(self, inputfiles):
        """ Runs several files from execute.
        """
        if len(inputfiles) > 0:
            # Read input files to datain
            self.datain = []
            for filename in inputfiles:
                # Read input file
                data = DataParent(config = self.config)
                self.datain.append(data.load(filename))
            # Call start - run and call end
            self.runstart(self.datain,self.arglist)
            self.run()
            self.runend(self.dataout)
            # Write output file
            self.dataout.save()
            self.log.info('Execute: Saved result %s' % self.dataout.filename)
        else:
            # Warning - no input file(s)
            self.log.warn('Execute: Missing input File(s)')

    def test(self):
        """ Test Pipe Step Parent Object:
            Runs a set of basic tests on the object
        """
        # log message
        self.log.info('Testing pipe step %s' %self.name)
        # read configuration
        if self.config != None:
            datain = DataParent(config=self.config)
        else:
            datain = DataParent(config=self.testconf)
        # generate 2 files
        datain.filename = 'this.file.type.fts'
        datain = [datain,datain]
        # run function call
        dataout = self(datain)
        # test output
        print(type(dataout))
        print(dataout.header)
        # log message
        self.log.info('Testing pipe step %s - Done' %self.name)

if __name__ == '__main__':
    """ Main function to run the pipe step from command line on a file.
        Command:
          python stepparent.py input.fits -arg1 -arg2 . . .
        Standard arguments:
          --config=ConfigFilePathName.txt : name of the configuration file
          -t, --test : runs the functionality test i.e. pipestep.test()
          --loglevel=LEVEL : configures the logging output for a particular level
          -h, --help : Returns a list of
    """
    StepMIParent().execute()

""" === History ===
    2014-3-14 Marc Berthoud: Written and tested version 0.1
"""
